"""Metrics to investigate quantities related to SRD.
Potentially could diverge from versions in scienceRadar.
"""

__all__ = ("fOBatch", "astrometryBatch", "rapidRevisitBatch")

import warnings

import healpy as hp
import numpy as np

import rubin_sim.maf.metric_bundles as mb
import rubin_sim.maf.metrics as metrics
import rubin_sim.maf.plots as plots
import rubin_sim.maf.slicers as slicers
import rubin_sim.maf.stackers as stackers

from .col_map_dict import col_map_dict
from .common import standard_summary


def fOBatch(
    colmap=None,
    run_name="run_name",
    extra_sql=None,
    extra_info=None,
    slicer=None,
    benchmark_area=18000,
    benchmark_n_visits=825,
    min_n_visits=750,
):
    """Metrics for calculating fO.

    Parameters
    ----------
    colmap : `dict` or None, opt
        A dictionary with a mapping of column names.
    run_name : `str`, opt
        The name of the simulated survey.
    extra_sql : `str` or None, opt
        Additional sql constraint to apply to all metrics.
    extra_Info : `str` or None, opt
        Additional info_label to apply to all results.
    slicer : `rubin_sim.maf.slicer.HealpixSlicer` or None, opt
         This must be a HealpixSlicer or some kind,
         although could be a HealpixSubsetSlicer.
         None will default to HealpixSlicer with nside=64.
    benchmark_area : `float`, opt
        Area to use when calculating fO_Nvis, for design.
    benchmark_n_visits : `float`, opt
        Nvisits minimum to use when calculating fO_Area, for design.
    min_n_visits : `float`, opt
        Nvisits minimum to use when calculating fO_Area, for minimum.

    Returns
    -------
    metric_bundleDict : `dict` of `maf.MetricBundle`
    """
    if colmap is None:
        colmap = col_map_dict()

    bundleList = []

    sql = ""
    info_label = "All visits"
    # Add additional sql constraint (such as wfdWhere) and info_label
    if (extra_sql is not None) and (len(extra_sql) > 0):
        sql = extra_sql
        if extra_info is None:
            info_label = extra_sql.replace("filter =", "").replace("filter=", "")
            info_label = info_label.replace('"', "").replace("'", "")
    if extra_info is not None:
        info_label = extra_info

    subgroup = info_label

    raCol = colmap["ra"]
    decCol = colmap["dec"]
    degrees = colmap["raDecDeg"]
    # Set up fO metric.
    if slicer is None:
        nside = 64
        slicer = slicers.HealpixSlicer(nside=nside, lat_col=decCol, lon_col=raCol, lat_lon_deg=degrees)
    else:
        try:
            nside = slicer.nside
        except AttributeError:
            warnings.warn("Must use a healpix slicer. Swapping to the default.")
            nside = 64
            slicer = slicers.HealpixSlicer(nside=nside, lat_col=decCol, lon_col=raCol, lat_lon_deg=degrees)

    displayDict = {"group": "SRD FO metrics", "subgroup": subgroup, "order": 0}

    # Configure the count metric which is what is used for f0 slicer.
    metric = metrics.CountExplimMetric(metric_name="fO", exp_col=colmap["exptime"])
    plotDict = {
        "xlabel": "Number of Visits",
        "asky": benchmark_area,
        "n_visits": min_n_visits,
        "x_min": 0,
        "x_max": 1500,
    }
    summaryMetrics = [
        metrics.FOArea(
            nside=nside,
            norm=False,
            metric_name="fOArea",
            asky=benchmark_area,
            n_visit=benchmark_n_visits,
        ),
        metrics.FOArea(
            nside=nside,
            norm=True,
            metric_name="fOArea/benchmark",
            asky=benchmark_area,
            n_visit=benchmark_n_visits,
        ),
        metrics.FONv(
            nside=nside,
            norm=False,
            metric_name="fONv",
            asky=benchmark_area,
            n_visit=benchmark_n_visits,
        ),
        metrics.FONv(
            nside=nside,
            norm=True,
            metric_name="fONv/benchmark",
            asky=benchmark_area,
            n_visit=benchmark_n_visits,
        ),
        metrics.FOArea(
            nside=nside,
            norm=False,
            metric_name=f"fOArea_{min_n_visits}",
            asky=benchmark_area,
            n_visit=min_n_visits,
        ),
    ]
    caption = "The FO metric evaluates the overall efficiency of observing. "
    caption += (
        "foNv: out of %.2f sq degrees, the area receives at least X and a median of Y visits "
        "(out of %d, if compared to benchmark). " % (benchmark_area, benchmark_n_visits)
    )
    caption += (
        "fOArea: this many sq deg (out of %.2f sq deg if compared "
        "to benchmark) receives at least %d visits. " % (benchmark_area, benchmark_n_visits)
    )
    displayDict["caption"] = caption
    bundle = mb.MetricBundle(
        metric,
        slicer,
        sql,
        plot_dict=plotDict,
        display_dict=displayDict,
        summary_metrics=summaryMetrics,
        plot_funcs=[plots.FOPlot()],
        info_label=info_label,
    )
    bundleList.append(bundle)
    # Set the run_name for all bundles and return the bundleDict.
    for b in bundleList:
        b.set_run_name(run_name)
    return mb.make_bundles_dict_from_list(bundleList)


def astrometryBatch(
    colmap=None,
    run_name="opsim",
    extra_sql=None,
    extra_info=None,
    slicer=None,
):
    """Metrics for evaluating proper motion and parallax.

    Parameters
    ----------
    colmap : `dict` or None, optional
        A dictionary with a mapping of column names.
    run_name : `str`, optional
        The name of the simulated survey.
    extra_sql : `str` or None, optional
        Additional sql constraint to apply to all metrics.
    extra_info : `str` or None, optional
        Additional info_label to apply to all results.
    slicer : `rubin_sim.maf.slicer` or None, optional
        Optionally, specify something other than an nside=64 healpix slicer.

    Returns
    -------
    metric_bundleDict : `dict` of `maf.MetricBundle`
    """
    if colmap is None:
        colmap = col_map_dict()
    bundleList = []

    sql = ""
    info_label = "All visits"
    # Add additional sql constraint (such as wfdWhere) and info_label
    if (extra_sql is not None) and (len(extra_sql) > 0):
        sql = extra_sql
        if extra_info is None:
            info_label = extra_sql.replace("filter =", "").replace("filter=", "")
            info_label = info_label.replace('"', "").replace("'", "")
    if extra_info is not None:
        info_label = extra_info

    subgroup = info_label

    raCol = colmap["ra"]
    decCol = colmap["dec"]
    degrees = colmap["raDecDeg"]

    rmags_para = [22.4, 24.0]
    rmags_pm = [20.5, 24.0]

    # Set up parallax/dcr stackers.
    parallaxStacker = stackers.ParallaxFactorStacker(
        ra_col=raCol, dec_col=decCol, date_col=colmap["mjd"], degrees=degrees
    )
    dcrStacker = stackers.DcrStacker(
        filter_col=colmap["filter"],
        alt_col=colmap["alt"],
        degrees=degrees,
        ra_col=raCol,
        dec_col=decCol,
        lst_col=colmap["lst"],
        site="LSST",
        mjd_col=colmap["mjd"],
    )

    # Set up parallax metrics.
    if slicer is None:
        slicer = slicers.HealpixSlicer(nside=64, lon_col=raCol, lat_col=decCol, lat_lon_deg=degrees)
    subsetPlots = [plots.HealpixSkyMap(), plots.HealpixHistogram()]

    displayDict = {
        "group": "SRD Parallax",
        "subgroup": subgroup,
        "order": 0,
        "caption": None,
    }
    # Expected error on parallax at 10 AU.
    plotmaxVals = (5.0, 18.0)
    good_parallax_limit = 11.5
    summary = [
        metrics.AreaSummaryMetric(
            area=18000,
            reduce_func=np.median,
            decreasing=False,
            metric_name="Median Parallax Uncert (18k)",
        ),
        metrics.AreaThresholdMetric(
            upper_threshold=good_parallax_limit,
            metric_name="Area better than %.1f mas uncertainty" % good_parallax_limit,
        ),
    ]
    summary.append(metrics.PercentileMetric(percentile=95, metric_name="95th Percentile Parallax Uncert"))
    summary.extend(standard_summary())
    for rmag, plotmax in zip(rmags_para, plotmaxVals):
        plotDict = {"x_min": 0, "x_max": plotmax, "color_min": 0, "color_max": plotmax}
        metric = metrics.ParallaxMetric(
            metric_name="Parallax Uncert @ %.1f" % (rmag),
            rmag=rmag,
            seeing_col=colmap["seeingGeom"],
            filter_col=colmap["filter"],
            m5_col=colmap["fiveSigmaDepth"],
            normalize=False,
        )
        bundle = mb.MetricBundle(
            metric,
            slicer,
            sql,
            info_label=info_label,
            stacker_list=[parallaxStacker],
            display_dict=displayDict,
            plot_dict=plotDict,
            summary_metrics=summary,
            plot_funcs=subsetPlots,
        )
        bundleList.append(bundle)
        displayDict["order"] += 1

    # Parallax normalized to 'best possible'
    # This separates the effect of cadence from depth.
    for rmag in rmags_para:
        metric = metrics.ParallaxMetric(
            metric_name="Normalized Parallax Uncert @ %.1f" % (rmag),
            rmag=rmag,
            seeing_col=colmap["seeingGeom"],
            filter_col=colmap["filter"],
            m5_col=colmap["fiveSigmaDepth"],
            normalize=True,
        )
        bundle = mb.MetricBundle(
            metric,
            slicer,
            sql,
            info_label=info_label,
            stacker_list=[parallaxStacker],
            display_dict=displayDict,
            summary_metrics=standard_summary(),
            plot_funcs=subsetPlots,
        )
        bundleList.append(bundle)
        displayDict["order"] += 1
    # Parallax factor coverage.
    for rmag in rmags_para:
        metric = metrics.ParallaxCoverageMetric(
            metric_name="Parallax Coverage @ %.1f" % (rmag),
            rmag=rmag,
            m5_col=colmap["fiveSigmaDepth"],
            mjd_col=colmap["mjd"],
            filter_col=colmap["filter"],
            seeing_col=colmap["seeingGeom"],
        )
        bundle = mb.MetricBundle(
            metric,
            slicer,
            sql,
            info_label=info_label,
            stacker_list=[parallaxStacker],
            display_dict=displayDict,
            summary_metrics=standard_summary(),
            plot_funcs=subsetPlots,
        )
        bundleList.append(bundle)
        displayDict["order"] += 1
    # Parallax problems can be caused by HA and DCR degeneracies.
    # Check their correlation.
    for rmag in rmags_para:
        metric = metrics.ParallaxDcrDegenMetric(
            metric_name="Parallax-DCR degeneracy @ %.1f" % (rmag),
            rmag=rmag,
            seeing_col=colmap["seeingEff"],
            filter_col=colmap["filter"],
            m5_col=colmap["fiveSigmaDepth"],
        )
        caption = "Correlation between parallax offset magnitude and hour angle for a r=%.1f star." % (rmag)
        caption += " (0 is good, near -1 or 1 is bad)."
        bundle = mb.MetricBundle(
            metric,
            slicer,
            sql,
            info_label=info_label,
            stacker_list=[dcrStacker, parallaxStacker],
            display_dict=displayDict,
            summary_metrics=standard_summary(),
            plot_funcs=subsetPlots,
        )
        bundleList.append(bundle)
        displayDict["order"] += 1

    # Evaluate y-band-only parallax uncertainty
    # Approximate "10sigma sources" as y=21.33
    ymag = 21.33
    if info_label == "All visits":
        yinfo = "y band visits"
    else:
        yinfo = f"{info_label} y band only"
    if len(sql) == 0:
        ysql = "filter == 'y'"
    else:
        ysql = f"{sql} and filter == 'y'"
    plotDict = {"x_min": 0, "x_max": 15, "color_min": 0, "color_max": 15}
    metric = metrics.ParallaxMetric(
        metric_name="Parallax Uncert @ %.1f" % (ymag),
        rmag=ymag,
        seeing_col=colmap["seeingGeom"],
        filter_col=colmap["filter"],
        m5_col=colmap["fiveSigmaDepth"],
        normalize=False,
    )
    bundle = mb.MetricBundle(
        metric,
        slicer,
        ysql,
        info_label=yinfo,
        stacker_list=[parallaxStacker],
        display_dict=displayDict,
        plot_dict=plotDict,
        summary_metrics=summary,
        plot_funcs=subsetPlots,
    )
    bundleList.append(bundle)
    displayDict["order"] += 1
    metric = metrics.ParallaxMetric(
        metric_name="Normalized Parallax Uncert @ %.1f" % (ymag),
        rmag=ymag,
        seeing_col=colmap["seeingGeom"],
        filter_col=colmap["filter"],
        m5_col=colmap["fiveSigmaDepth"],
        normalize=True,
    )
    bundle = mb.MetricBundle(
        metric,
        slicer,
        ysql,
        info_label=yinfo,
        stacker_list=[parallaxStacker],
        display_dict=displayDict,
        summary_metrics=summary,
        plot_funcs=subsetPlots,
    )
    bundleList.append(bundle)
    displayDict["order"] += 1

    # Proper Motion metrics.
    displayDict = {
        "group": "SRD Proper Motion",
        "subgroup": subgroup,
        "order": 0,
        "caption": None,
    }
    # Proper motion errors.
    plotmaxVals = (1.0, 5.0)
    summary = [
        metrics.AreaSummaryMetric(
            area=18000,
            reduce_func=np.median,
            decreasing=False,
            metric_name="Median Proper Motion Uncert (18k)",
        )
    ]
    summary.append(metrics.PercentileMetric(metric_name="95th Percentile Proper Motion Uncert"))
    summary.extend(standard_summary())
    for rmag, plotmax in zip(rmags_pm, plotmaxVals):
        plotDict = {"x_min": 0, "x_max": plotmax, "color_min": 0, "color_max": plotmax}
        metric = metrics.ProperMotionMetric(
            metric_name="Proper Motion Uncert @ %.1f" % rmag,
            rmag=rmag,
            m5_col=colmap["fiveSigmaDepth"],
            mjd_col=colmap["mjd"],
            filter_col=colmap["filter"],
            seeing_col=colmap["seeingGeom"],
            normalize=False,
        )
        bundle = mb.MetricBundle(
            metric,
            slicer,
            sql,
            info_label=info_label,
            display_dict=displayDict,
            plot_dict=plotDict,
            summary_metrics=summary,
            plot_funcs=subsetPlots,
        )
        bundleList.append(bundle)
        displayDict["order"] += 1
    # Normalized proper motion.
    for rmag in rmags_pm:
        metric = metrics.ProperMotionMetric(
            metric_name="Normalized Proper Motion Uncert @ %.1f" % rmag,
            rmag=rmag,
            m5_col=colmap["fiveSigmaDepth"],
            mjd_col=colmap["mjd"],
            filter_col=colmap["filter"],
            seeing_col=colmap["seeingGeom"],
            normalize=True,
        )
        bundle = mb.MetricBundle(
            metric,
            slicer,
            sql,
            info_label=info_label,
            display_dict=displayDict,
            summary_metrics=standard_summary(),
            plot_funcs=subsetPlots,
        )
        bundleList.append(bundle)
        displayDict["order"] += 1

    # Set the run_name for all bundles and return the bundleDict.
    for b in bundleList:
        b.set_run_name(run_name)
    return mb.make_bundles_dict_from_list(bundleList)


def rapidRevisitBatch(
    colmap=None,
    run_name="opsim",
    extra_sql=None,
    extra_info=None,
    slicer=None,
):
    """Metrics for evaluating proper motion and parallax.

    Parameters
    ----------
    colmap : `dict` or None, optional
        A dictionary with a mapping of column names.
    run_name : `str`, optional
        The name of the simulated survey.
    extra_sql : `str` or None, optional
        Additional sql constraint to apply to all metrics.
    extra_info : `str` or None, optional
        Additional info_label to apply to all results.
    slicer : `rubin_sim_maf.slicers.HealpixSlicer` or None, optional
        Optionally, specify something other than an nside=64 healpix slicer.
        (must be a healpix slicer)

    Returns
    -------
    metric_bundleDict : `dict` of `maf.MetricBundle`
    """
    if colmap is None:
        colmap = col_map_dict()
    bundleList = []

    sql = ""
    info_label = "All visits"
    # Add additional sql constraint (such as wfdWhere) and info_label.
    if (extra_sql is not None) and (len(extra_sql) > 0):
        sql = extra_sql
        if extra_info is None:
            info_label = extra_sql.replace("filter =", "").replace("filter=", "")
            info_label = info_label.replace('"', "").replace("'", "")
    if extra_info is not None:
        info_label = extra_info

    subgroup = info_label

    raCol = colmap["ra"]
    decCol = colmap["dec"]
    degrees = colmap["raDecDeg"]

    if slicer is None:
        nside = 64
        slicer = slicers.HealpixSlicer(nside=nside, lon_col=raCol, lat_col=decCol, lat_lon_deg=degrees)
    else:
        try:
            nside = slicer.nside
        except AttributeError:
            warnings.warn("Must use a healpix slicer. Swapping to the default.")
            nside = 64
            slicer = slicers.HealpixSlicer(nside=nside, lat_col=decCol, lon_col=raCol, lat_lon_deg=degrees)

    subsetPlots = [plots.HealpixSkyMap(), plots.HealpixHistogram()]

    displayDict = {
        "group": "SRD Rapid Revisits",
        "subgroup": subgroup,
        "order": 0,
        "caption": None,
    }

    # Calculate the actual number of revisits within 30 minutes.
    dTmax = 30  # time in minutes
    m2 = metrics.NRevisitsMetric(
        d_t=dTmax,
        mjd_col=colmap["mjd"],
        normed=False,
        metric_name="NumberOfQuickRevisits",
    )
    plotDict = {"color_min": 400, "color_max": 2000, "x_min": 400, "x_max": 2000}
    caption = "Number of consecutive visits with return times faster than %.1f minutes, " % (dTmax)
    caption += "in any filter, all proposals. "
    displayDict["caption"] = caption
    bundle = mb.MetricBundle(
        m2,
        slicer,
        sql,
        plot_dict=plotDict,
        plot_funcs=subsetPlots,
        info_label=info_label,
        display_dict=displayDict,
        summary_metrics=standard_summary(with_count=False),
    )
    bundleList.append(bundle)
    displayDict["order"] += 1

    # Better version of the rapid revisit requirements:
    # require a minimum number of visits between
    # dtMin and dtMax, but also a minimum number of visits
    # between dtMin and dtPair (the typical pair time).
    # 1 means the healpix met the requirements (0 means did not).
    dTmin = 40.0 / 60.0  # (minutes) 40s minimum for rapid revisit range
    dTpairs = 20.0  # minutes (time when pairs should start kicking in)
    dTmax = 30.0  # 30 minute maximum for rapid revisit range
    nOne = 82  # Number of revisits between 40s-30m required
    nTwo = 28  # Number of revisits between 40s - tPairs required.
    pix_area = float(hp.nside2pixarea(nside, degrees=True))
    scale = pix_area * hp.nside2npix(nside)
    m1 = metrics.RapidRevisitMetric(
        metric_name="RapidRevisits",
        mjd_col=colmap["mjd"],
        d_tmin=dTmin / 60.0 / 60.0 / 24.0,
        d_tpairs=dTpairs / 60.0 / 24.0,
        d_tmax=dTmax / 60.0 / 24.0,
        min_n1=nOne,
        min_n2=nTwo,
    )
    plotDict = {
        "x_min": 0,
        "x_max": 1,
        "color_min": 0,
        "color_max": 1,
        "log_scale": False,
    }
    cutoff1 = 0.9
    summaryStats = [metrics.FracAboveMetric(cutoff=cutoff1, scale=scale, metric_name="Area (sq deg)")]
    caption = "Rapid Revisit: area that receives at least %d visits between %.3f and %.1f minutes, " % (
        nOne,
        dTmin,
        dTmax,
    )
    caption += "with at least %d of those visits falling between %.3f and %.1f minutes. " % (
        nTwo,
        dTmin,
        dTpairs,
    )
    caption += (
        'Summary statistic "Area" indicates the area on the sky which meets this requirement.'
        " (SRD design specification is 2000 sq deg)."
    )
    displayDict["caption"] = caption
    bundle = mb.MetricBundle(
        m1,
        slicer,
        sql,
        plot_dict=plotDict,
        plot_funcs=subsetPlots,
        info_label=info_label,
        display_dict=displayDict,
        summary_metrics=summaryStats,
    )
    bundleList.append(bundle)
    displayDict["order"] += 1

    # Set the run_name for all bundles and return the bundleDict.
    for b in bundleList:
        b.set_run_name(run_name)
    return mb.make_bundles_dict_from_list(bundleList)



"""Sets of metrics to look at impact of cadence on science
"""

__all__ = ("phaseGap",)

import rubin_sim.maf.metric_bundles as mb
import rubin_sim.maf.metrics as metrics
import rubin_sim.maf.plots as plots
import rubin_sim.maf.slicers as slicers

from .col_map_dict import col_map_dict
from .common import combine_info_labels, standard_summary


def phaseGap(
    colmap=None,
    runName="opsim",
    nside=64,
    extraSql=None,
    extraInfoLabel=None,
):
    """Generate a set of statistics about period coverage and phase gaps.

    Parameters
    ----------
    colmap : `dict` or None, optional
        A dictionary with a mapping of column names.
    runName : `str`, optional
        The name of the simulated survey.
    nside : `int`, optional
        Nside for the healpix slicer.
    extraSql : `str` or None, optional
        Additional sql constraint to apply to all metrics.
    extraInfoLabel : `str` or None, optional
        Additional info_label to apply to all results.

    Returns
    -------
    metric_bundleDict : `dict` of `maf.MetricBundle`
    """

    if colmap is None:
        colmap = col_map_dict()

    info_label = extraInfoLabel
    if extraSql is not None and len(extraSql) > 0:
        if info_label is None:
            info_label = extraSql

    raCol = colmap["ra"]
    decCol = colmap["dec"]
    degrees = colmap["raDecDeg"]

    bundleList = []
    standardStats = standard_summary()
    subsetPlots = [plots.HealpixSkyMap(), plots.HealpixHistogram()]

    slicer = slicers.HealpixSlicer(nside=nside, lat_col=decCol, lon_col=raCol, lat_lon_deg=degrees)

    # largest phase gap for periods
    periods = [0.1, 1.0, 10.0, 100.0]
    sqls = [
        'filter = "u"',
        'filter="r"',
        'filter="g" or filter="r" or filter="i" or filter="z"',
        "",
    ]
    filter_names = ["u", "r", "griz", "all"]
    info_labels = filter_names
    if info_label is not None:
        info_labels = [combine_info_labels(m, info_label) for m in info_labels]
    if extraSql is not None and len(extraSql) > 0:
        for sql in sqls:
            sqls[sql] = "(%s) and (%s)" % (sqls[sql], extraSql)

    for sql, md, f in zip(sqls, info_labels, filter_names):
        for period in periods:
            displayDict = {
                "group": "PhaseGap",
                "subgroup": "Filter %s: Period %.2f days" % (f, period),
                "caption": "Maximum phase gap, given a period of %.2f days." % period,
            }
            metric = metrics.PhaseGapMetric(
                nPeriods=1,
                periodMin=period,
                periodMax=period,
                nVisitsMin=5,
                metric_name="PhaseGap %.1f day" % period,
            )
            metric.reduceFuncs = {metric.reduceFuncs["reduceLargestGap"]}
            metric.reduceOrder = {0}
            bundle = mb.MetricBundle(
                metric,
                slicer,
                constraint=sql,
                info_label=md,
                display_dict=displayDict,
                summary_metrics=standardStats,
                plot_funcs=subsetPlots,
            )
            bundleList.append(bundle)

    # Set the run_name for all bundles and return the bundleDict.
    for b in bundleList:
        b.set_run_name(runName)
    return mb.make_bundles_dict_from_list(bundleList)



__all__ = ("PhotometricSelfCalUniformityMetric",)

import os
import time

import healpy as hp
import numpy as np
from astropy.coordinates import SkyCoord
from astropy.table import Table
from rubin_scheduler.data import get_data_dir
from rubin_scheduler.utils import healbin
from scipy.stats import median_abs_deviation

from rubin_sim.maf.metrics import BaseMetric
from rubin_sim.selfcal import LsqrSolver, OffsetSNR, generate_catalog
from rubin_sim.selfcal.offsets import OffsetSys


def _match(arr1, arr2):
    st1 = np.argsort(arr1)
    sub1 = np.searchsorted(arr1, arr2, sorter=st1)
    if arr2.max() > arr1.max():
        bad = sub1 == arr1.size
        sub1[bad] = arr1.size - 1

    (sub2,) = np.where(arr1[st1[sub1]] == arr2)
    sub1 = st1[sub1[sub2]]

    return sub1, sub2


class PhotometricSelfCalUniformityMetric(BaseMetric):
    def __init__(
        self,
        nside_residual=128,
        highglat_cut=30.0,
        outlier_nsig=4.0,
        metric_name="PhotometricSelfCalUniformityMetric",
        filter_name="r",
    ):
        cols = [
            "observationid",
            "fieldra",
            "fielddec",
            "fiveSigmaDepth",
            "rotSkyPos",
            "filter",
        ]
        super().__init__(col=cols, metric_name=metric_name)

        filename = os.path.join(get_data_dir(), "maf", "monster_stars_uniformity_i15-18_sampled.parquet")
        self.stars = Table.read(filename)

        # We have to rename dec to decl for the selfcal code.
        if "dec" in self.stars.dtype.names:
            self.stars["decl"] = self.stars["dec"]

        self.stars = self.stars.as_array()

        self.nside_residual = nside_residual
        self.highglat_cut = highglat_cut
        self.outlier_nsig = outlier_nsig

        self.units = "mmag"
        self.filter_name = filter_name

    def run(self, data_slice, slice_point=None):
        offsets = [OffsetSys(error_sys=0.03), OffsetSNR(lsst_filter=self.filter_name)]

        visits = np.zeros(
            len(data_slice),
            dtype=[
                ("observationId", "i8"),
                ("ra", "f8"),
                ("dec", "f8"),
                ("fiveSigmaDepth", "f8"),
                ("rotSkyPos", "f8"),
            ],
        )
        visits["observationId"] = data_slice["observationId"]
        visits["ra"] = data_slice["fieldRA"]
        visits["dec"] = data_slice["fieldDec"]
        visits["fiveSigmaDepth"] = data_slice["fiveSigmaDepth"]
        visits["rotSkyPos"] = data_slice["rotSkyPos"]

        good_stars = np.isfinite(self.stars[f"{self.filter_name}mag"])
        stars = self.stars[good_stars]

        observed_stars = generate_catalog(
            visits,
            stars,
            offsets=offsets,
            lsst_filter=self.filter_name,
            n_patches=16,
            verbose=False,
        )

        solver = LsqrSolver(observed_stars)

        print("Starting solver...")
        t0 = time.time()

        solver.run()
        fit_patches, fit_stars = solver.return_solution()

        t1 = time.time()
        dt = (t1 - t0) / 60.0
        print("runtime= %.1f min" % dt)

        # Trim stars to the ones with solutions.
        a, b = _match(stars["id"], fit_stars["id"])
        stars_trimmed = stars[a]
        fit_stars_trimmed = fit_stars[b]

        # Residuals after fit, removing floating zeropoint
        resid = stars_trimmed[f"{self.filter_name}mag"] - fit_stars_trimmed["fit_mag"]
        resid = resid - np.median(resid)

        resid_map = healbin(
            stars_trimmed["ra"],
            stars_trimmed["dec"],
            resid,
            self.nside_residual,
            reduce_func=np.median,
        )

        (ipring,) = np.where(resid_map > hp.UNSEEN)

        scatter_full = median_abs_deviation(resid_map[ipring], scale="normal")

        # Fraction of (4) sigma outliers.
        highscat = np.abs(resid_map[ipring]) > self.outlier_nsig * scatter_full
        outlier_frac_full = highscat.sum() / len(ipring)

        # Cut to high latitude
        ra, dec = hp.pix2ang(self.nside_residual, ipring, nest=False, lonlat=True)
        coords = SkyCoord(ra, dec, frame="icrs", unit="deg")
        b = coords.galactic.b.value

        high_glat = np.abs(b) > self.highglat_cut
        scatter_highglat = median_abs_deviation(resid_map[ipring[high_glat]], scale="normal")

        # Fraction of (4) sigma outliers.
        highscat = np.abs(resid_map[ipring[high_glat]]) > self.outlier_nsig * scatter_highglat
        outlier_frac_highglat = highscat.sum() / len(ipring)

        # Convert to mmag
        scatter_full = scatter_full * 1000.0
        scatter_highglat = scatter_highglat * 1000.0

        result = {
            "scatter_full": scatter_full,
            "scatter_highglat": scatter_highglat,
            "outlier_frac_full": outlier_frac_full,
            "outlier_frac_highglat": outlier_frac_highglat,
            "uniformity_map": resid_map,
        }

        return result

    def reduce_scatter_full(self, metric_value):
        return metric_value["scatter_full"]

    def reduce_scatter_highglat(self, metric_value):
        return metric_value["scatter_highglat"]

    def reduce_outlier_frac_full(self, metric_value):
        return metric_value["outlier_frac_full"]

    def reduce_outlier_frac_highglat(self, metric_value):
        return metric_value["outlier_frac_highglat"]



# Various things declared here to be imported when running
# artificialStructureCalculation.
# Makes the updates easier, since the constants/objects defined here
# are used by different
# functions, e.g. power law constants are called by
# GalaxyCountsMetric_extended as well as
# GalaxyCounts_withPixelCalibration.
#
# Included here:
#  * power law constants from the i-band mock catalog; based on mocks
#  from Padilla et al.
#  * normalization constant for galaxy counts from the mock catalogs.
#
# Humna Awan
# humna.awan@rutgers.edu

# Power law constants for each z-bin based on N. D. Padilla et al.'s
# mock catalogs
# General power law form: 10**(a*m+b)
# Declare the dictionary for the power law constants
from collections import OrderedDict

power_law_const_a = OrderedDict()
power_law_const_b = OrderedDict()
# 0.<z<0.15
power_law_const_a["0.<z<0.15"] = 0.166  # July 18 Catalog
power_law_const_b["0.<z<0.15"] = -0.989  # July 18 Catalog

# 0.15<z<0.37
power_law_const_a["0.15<z<0.37"] = 0.212  # July 18 Catalog
power_law_const_b["0.15<z<0.37"] = -1.540  # July 18 Catalog

# 0.37<z<0.66
power_law_const_a[" 0.37<z<0.66"] = 0.250  # July 18 Catalog
power_law_const_b[" 0.37<z<0.66"] = -2.242  # July 18 Catalog

# 0.66<z<1.0
power_law_const_a["0.66<z<1.0"] = 0.265  # July 18 Catalog
power_law_const_b["0.66<z<1.0"] = -2.660  # July 18 Catalog

# 1.0<z<1.5
power_law_const_a["1.0<z<1.5"] = 0.283  # July 18 Catalog
power_law_const_b["1.0<z<1.5"] = -3.242  # July 18 Catalog

# 1.5<z<2.0
power_law_const_a["1.5<z<2.0"] = 0.308  # July 18 Catalog
power_law_const_b["1.5<z<2.0"] = -4.169  # July 18 Catalog

# 2.0<z<2.5
power_law_const_a["2.0<z<2.5"] = 0.364  # July 18 Catalog
power_law_const_b["2.0<z<2.5"] = -6.261  # July 18 Catalog

# 2.5<z<3.0
power_law_const_a["2.5<z<3.0"] = 0.346  # July 18 Catalog
power_law_const_b["2.5<z<3.0"] = -5.661  # July 18 Catalog

# 3.0<z<3.5
power_law_const_a["3.0<z<3.5"] = 0.381  # July 18 Catalog
power_law_const_b["3.0<z<3.5"] = -6.810  # July 18 Catalog

# 3.5<z<4.0
power_law_const_a["3.5<z<4.0"] = 0.424  # July 18 Catalog
power_law_const_b["3.5<z<4.0"] = -8.362  # July 18 Catalog

# 4.0<z<15.0
power_law_const_a["4.0<z<15.0"] = 0.462  # July 18 Catalog
power_law_const_b["4.0<z<15.0"] = -9.474  # July 18 Catalog


# Normalization to match the total galaxies from  mock catalogs to match
# up to CFHTLS counts for
# i<25.5 galaxy catalog. Using July 18-19 i-band mock catalogs.
normalization_constant = 3.70330701802



__all__ = (
    "generate_known_lv_dwarf_slicer",
    "make__fake_old_galaxy_lf",
    "make_dwarf_lf_dicts",
    "LVDwarfsMetric",
)

import os

import astropy.units as u
import numpy as np
from astropy.coordinates import SkyCoord
from astropy.io import ascii, fits
from rubin_scheduler.data import get_data_dir

from rubin_sim.maf.maf_contrib.lss_obs_strategy import GalaxyCountsMetricExtended
from rubin_sim.maf.metrics import BaseMetric, ExgalM5, StarDensityMetric
from rubin_sim.maf.slicers import UserPointsSlicer


def generate_known_lv_dwarf_slicer():
    """Read the Karachentsev+ catalog of nearby galaxies,
    and put the info about them into a UserPointSlicer object.
    """
    filename = os.path.join(get_data_dir(), "maf/lvdwarfs", "lsst_galaxies_1p25to9Mpc_table.fits")
    lv_dat0 = fits.getdata(filename)

    # Keep only galaxies at dec < 35 deg.,
    # and with stellar masses > 10^7 M_Sun (and <1e14).
    lv_dat_cuts = (lv_dat0["dec"] < 35.0) & (lv_dat0["MStars"] > 1e7) & (lv_dat0["MStars"] < 1e14)
    lv_dat = lv_dat0[lv_dat_cuts]

    # Set up the slicer to evaluate the catalog we just made
    slicer = UserPointsSlicer(lv_dat["ra"], lv_dat["dec"], lat_lon_deg=True, badval=-666)
    # Add any additional information about each object to the slicer
    slicer.slice_points["distance"] = lv_dat["dist_Mpc"]

    return slicer


# make a simulated LF for old galaxy of given integrated B,
# distance modulus mu, in any of filters ugrizY
def make__fake_old_galaxy_lf(int_b, mu, filtername):
    """
    Make a simulated luminosity function for an old (10 Gyr) dwarf
    galaxy of given integrated B magnitude, at a given distance modulus,
    in any of the filters ugrizy.

    Parameters
    ----------
    int_b : `float`
        Integrated B-band magnitude of the dwarf to simulate.
    mu : `float`
        Distance modulus at which to place the simulated dwarf.
    filternams: `str`
        Filter in which to produce the simulated luminosity function.
    """
    if filtername == "y":
        filtername == "Y"
    model_bmag = 6.856379  # integrated B mag of the model LF being read
    # Read a simulated luminosity function of [M/H]=-1.5, 10 Gyr stellar
    # population:
    filename = os.path.join(get_data_dir(), "maf/lvdwarfs", "LF_-1.5_10Gyr.dat")
    LF = ascii.read(filename, header_start=12)
    mags = LF["magbinc"]
    counts = LF[filtername + "mag"]
    # shift model LF to requested distance and dim it
    mags = mags + mu
    model_bmag = model_bmag + mu
    # scale model counts up/down to reach the requested int_b
    factor = np.power(10.0, -0.4 * (int_b - model_bmag))
    counts = factor * counts
    return mags, counts


def make_dwarf_lf_dicts():
    """
    Create dicts containing g- and i-band LFs for simulated dwarfs between
    -10 < M_B < +3, so they can simply be looked up rather than having to
    recreate them each time. Dict is keyed on M_B value.
    """
    lf_dict_i = {}
    lf_dict_g = {}
    # Simulate a range from M_B=-10 to M_B=+5 in 0.1-mag increments.
    tmp_mb = -10.0

    for i in range(151):
        mbkey = f"MB{tmp_mb:.2f}"
        i_l_fmags, i_l_fcounts = make__fake_old_galaxy_lf(tmp_mb, 0.0, "i")
        lf_dict_i[mbkey] = (np.array(i_l_fmags), np.array(i_l_fcounts))
        g_l_fmags, g_l_fcounts = make__fake_old_galaxy_lf(tmp_mb, 0.0, "g")
        lf_dict_g[mbkey] = (np.array(g_l_fmags), np.array(g_l_fcounts))
        tmp_mb += 0.1

    return lf_dict_g, lf_dict_i


def _sum_luminosity(l_fmags, l_fcounts):
    """
    Sum the luminosities from a given luminosity function.

    Uses the first bin's magnitude as a reference, sums luminosities
    relative to that reference star,
    then converts back to magnitude at the end.

    Parameters
    ----------
    l_fmags : `np.array,` `float`
        Magnitude bin values from the simulated LF.
    l_fcounts : `np.array`, `int`
        Number of stars in each magnitude bin.
    """
    magref = l_fmags[0]
    totlum = 0.0

    for mag, count in zip(l_fmags, l_fcounts):
        tmpmags = np.repeat(mag, count)
        totlum += np.sum(10.0 ** ((magref - tmpmags) / 2.5))

    mtot = magref - 2.5 * np.log10(totlum)
    return mtot


def _dwarf_sblimit(glim, ilim, nstars, lf_dict_g, lf_dict_i, distlim, rng):
    """
    Calculate the surface brightness limit given the g- and i-band limiting
    magnitudes and the number of stars required to detect the dwarf of
    interest.

    Parameters
    ----------
    glim : `float`
        Limiting magnitude (coaddM5) in g-band.
    ilim : `float`
        Limiting magnitude (coaddM5) in i-band.
    nstars : `int`
        Number of stars required to be able to detect the dwarf.
    lf_dict_g, lf_dict_i : `dict`
        Dicts of computed luminosity functions for artificial dwarfs, as
        calculated by the function make_LF_dicts.
    distlim : `float`
        Distance limit (in Mpc) for which to calculate the limiting
        surface brightness for dwarf detection.
    rng : `np.random.Generator`
        Random noise generator for poisson random number use.
    """
    distance_limit = distlim * 1e6  # distance limit in parsecs
    distmod_lim = 5.0 * np.log10(distance_limit) - 5.0

    if (glim > 15) and (ilim > 15):
        # print(glim, ilim, nstars)
        fake_mb = -10.0
        ng = 1e6
        ni = 1e6

        while (ng > nstars) and (ni > nstars) and fake_mb < 5.0:
            # B_fake = distmod_limit+fake_mb
            mbkey = f"MB{fake_mb:.2f}"
            i_l_fmags0, i_l_fcounts0 = lf_dict_i[mbkey]
            g_l_fmags0, g_l_fcounts0 = lf_dict_g[mbkey]
            i_l_fcounts = rng.poisson(i_l_fcounts0)
            g_l_fcounts = rng.poisson(g_l_fcounts0)
            # Add the distance modulus to make it apparent mags
            i_l_fmags = i_l_fmags0 + distmod_lim
            # Add the distance modulus to make it apparent mags
            g_l_fmags = g_l_fmags0 + distmod_lim
            gsel = g_l_fmags <= glim
            isel = i_l_fmags <= ilim
            ng = np.sum(g_l_fcounts[gsel])
            ni = np.sum(i_l_fcounts[isel])
            fake_mb += 0.1

        if fake_mb > -9.9 and (ng > 0) and (ni > 0):
            gmag_tot = _sum_luminosity(g_l_fmags[gsel], g_l_fcounts[gsel]) - distmod_lim
            imag_tot = _sum_luminosity(i_l_fmags[isel], i_l_fcounts[isel]) - distmod_lim
            # S = m + 2.5logA, where in this case things are in sq. arcmin,
            # so A = 1 arcmin^2 = 3600 arcsec^2
            sbtot_g = distmod_lim + gmag_tot + 2.5 * np.log10(3600.0)
            sbtot_i = distmod_lim + imag_tot + 2.5 * np.log10(3600.0)
            mg_lim = gmag_tot
            mi_lim = imag_tot
            sbg_lim = sbtot_g
            sbi_lim = sbtot_i
            if ng < ni:
                flag_lim = "g"
            else:
                flag_lim = "i"
        else:
            mg_lim = 999.9
            mi_lim = 999.9
            sbg_lim = 999.9
            sbi_lim = 999.9
            flag_lim = None
    else:
        mg_lim = 999.9
        mi_lim = 999.9
        sbg_lim = -999.9
        sbi_lim = -999.9
        flag_lim = None

    return mg_lim, mi_lim, sbg_lim, sbi_lim, flag_lim


class LVDwarfsMetric(BaseMetric):
    """
    Estimate the detection limit in total dwarf luminosity for
    resolved dwarf galaxies at a given distance.

    This metric class uses simulated luminosity functions of dwarf galaxies
    with known (assumed) luminosities to estimate the detection limit
    (in total dwarf luminosity, M_V) for resolved dwarf galaxies at a
    given distance. It can be applied to either known galaxies with their
    discrete positions and distances, or an entire survey simulation with
    a fixed distance limit.

    In the default use (with the KnownLvDwarfsSlicer),
    it returns detection limits for a catalog of known local volume dwarfs,
    from the Karachentsev+ catalog of nearby galaxies.

    Parameters
    ----------
    radius : `float`
        Radius of the field being considered (for discrete fields only).
        By default, UserPointSlicer uses a 2.45-deg field radius.
    distlim : `float`
        Distance threshold in Mpc for which to calculate the limiting
        dwarf detection luminosity. Only needed for healpix slicers,
        but *required* if healpix is used.
    cmd_frac : `float`
        Fraction of the total area of the color-magnitude diagram
        that is spanned by the tracer selection criteria. (e.g.,
        the size of a box in color and magnitude to select RGB-star candidates)
    stargal_contamination : `float`
        Fraction of objects in CMD selection region that are actually
        unresolved galaxies that were mis-classified as stars.
    nsigma : `float`
        Required detection significance to declare a simulated
        dwarf "detected."
    """

    def __init__(
        self,
        radius=2.45,
        distlim=None,
        cmd_frac=0.1,
        stargal_contamination=0.40,
        nsigma=10.0,
        metric_name="LVDwarfs",
        seed=505,
        **kwargs,
    ):
        self.radius = radius
        self.filter_col = "filter"
        self.m5_col = "fiveSigmaDepth"
        self.cmd_frac = cmd_frac
        self.stargal_contamination = stargal_contamination
        self.nsigma = nsigma

        if distlim is not None:
            # If the distance limit was set on input, extract that info:
            self.distlim = distlim
        else:
            # If no distance limit specified, assume the intention is to search
            #   around known Local Volume host galaxies.
            self.distlim = None
            filename = os.path.join(get_data_dir(), "maf/lvdwarfs", "lsst_galaxies_1p25to9Mpc_table.fits")
            lv_dat0 = fits.getdata(filename)
            # Keep only galaxies at dec < 35 deg.,
            # and with stellar masses > 10^7 M_Sun.
            lv_dat_cuts = (lv_dat0["dec"] < 35.0) & (lv_dat0["MStars"] > 1e7) & (lv_dat0["MStars"] < 1e14)
            lv_dat = lv_dat0[lv_dat_cuts]
            sc_dat = SkyCoord(
                ra=lv_dat["ra"] * u.deg,
                dec=lv_dat["dec"] * u.deg,
                distance=lv_dat["dist_Mpc"] * u.Mpc,
            )
            self.sc_dat = sc_dat

        self.lf_dict_g, self.lf_dict_i = make_dwarf_lf_dicts()

        # Set up already-defined metrics that we will need:
        self.exgal_coaddm5 = ExgalM5(m5_col=self.m5_col, filter_col=self.filter_col)
        # The StarDensityMetric calculates the number of stars in i band
        self.star_density_metric = StarDensityMetric(filtername="i")
        # The galaxy counts metric calculates the number of galaxies in i band
        self.galaxy_counts_metric = GalaxyCountsMetricExtended(
            m5_col=self.m5_col, filter_band="i", include_dust_extinction=True
        )
        # Set the scale for the GalaxyCountMetric_extended to 1, so it returns
        # galaxies per sq deg, not n galaxies per healpixel
        self.galaxy_counts_metric.scale = 1

        cols = [self.m5_col, self.filter_col]
        # GalaxyCountsMetric needs the DustMap,
        # and StarDensityMetric needs StellarDensityMap:
        maps = ["DustMap", "StellarDensityMap"]
        super().__init__(col=cols, metric_name=metric_name, maps=maps, units="M_V limit", **kwargs)

        # Set up a random number generator,
        # so that metric results are repeatable
        self.rng = np.random.default_rng(seed)

    def run(self, data_slice, slice_point=None):
        # Identify observations in g and i bandpasses
        gband = data_slice[self.filter_col] == "g"
        iband = data_slice[self.filter_col] == "i"
        # if there are no visits in either of g or i band, exit
        if len(np.where(gband)[0]) == 0 or len(np.where(iband)[0]) == 0:
            return self.badval

        # calculate the dust-extincted coadded 5-sigma limiting mags
        # in the g and i bands:
        g5 = self.exgal_coaddm5.run(data_slice[gband], slice_point)
        i5 = self.exgal_coaddm5.run(data_slice[iband], slice_point)

        if g5 < 15 or i5 < 15:
            # If the limiting magnitudes won't even match the
            # stellar density maps, exit
            return self.badval

        # Find the number of stars per sq arcsecond at the i band limit
        # (this is a bit of a hack to get the starDensityMetric to
        # calculate the nstars at this mag exactly)
        star_i5 = min(27.9, i5)
        self.star_density_metric.magLimit = star_i5

        nstar_sqarcsec = self.star_density_metric.run(data_slice, slice_point)

        # Calculate the number of galaxies per sq degree
        ngal_sqdeg = self.galaxy_counts_metric.run(data_slice, slice_point)
        # GalaxyCountsMetric is undefined in some places. These cases return
        #   zero; catch these and set the galaxy counts in those regions to a
        #   very high value. (this may not be true after catching earlier
        #   no-visits issues)
        if ngal_sqdeg < 10.0:
            ngal_sqdeg = 1e7

        # Convert from per sq deg and per sq arcsecond
        # into #'s per sq arcminute
        ngal_sqarcmin = ngal_sqdeg / 3600
        nstar_sqarcmin = nstar_sqarcsec * 3600

        if ngal_sqarcmin < 0 or nstar_sqarcmin < 0:
            print(
                f"Here be a problem - ngals_sqarcmin {ngal_sqarcmin} or nstar_sqarcmin {nstar_sqarcmin} "
                f"are negative. depths: {g5}, {i5}. "
                f'{slice_point["ra"], slice_point["dec"], slice_point["sid"]}'
            )
        # The number of stars required to reach nsigma is
        # nsigma times the Poisson fluctuations of the background
        # (stars+galaxies contamination):
        nstars_required = self.nsigma * np.sqrt(
            (ngal_sqarcmin * self.cmd_frac * self.stargal_contamination) + (nstar_sqarcmin * self.cmd_frac)
        )

        if self.distlim is not None:
            # Use the distlim if a healpix slicer is input
            distlim = self.distlim
        else:
            # Use discrete distances for known galaxies if a
            # UserPointSlicer:
            distlim = slice_point["distance"] * u.Mpc

        # Calculate the limiting luminosity and surface brightness
        # based on g5 and i5:
        mg_lim, mi_lim, sb_g_lim, sb_i_lim, flag_lim = _dwarf_sblimit(
            g5,
            i5,
            nstars_required,
            self.lf_dict_g,
            self.lf_dict_i,
            distlim=distlim.value,
            rng=self.rng,
        )

        if flag_lim is None:
            mv = self.badval

        else:
            # To go from HSC g and i bands to V, use the conversion
            # from Appendix A of Komiyama+2018, ApJ, 853, 29:
            # V = g_hsc - 0.371*(gi_hsc)-0.068
            mv = mg_lim - 0.371 * (mg_lim - mi_lim) - 0.068
            # sbv = sb_g_lim - 0.371 * (sb_g_lim - sb_i_lim) - 0.068

        return mv



#!/usr/bin/env python

# Mike Lund - Vanderbilt University
# mike.lund@gmail.com
# Last edited 8/15/2015
# Description: Calculates the stellar density based off of
# Juric et al 2008 and Jackson et al 2002. For use with Field Star Count metric

import numpy as np

zsun = 25.0
rsun = 8000.0
density_rsun = 0.0364
f = 0.1


def diskprofile(R, Z, L, H):
    part1 = (-R / L) - (abs(Z + zsun) / H)
    part2 = np.exp(part1)
    part3 = np.exp(rsun / L)
    tot = density_rsun * part2 * part3
    return tot


def thindisk(R, Z):
    return diskprofile(R, Z, 2150.0, 245.0)


def thickdisk(R, Z):
    return diskprofile(R, Z, 3261.0, 743.0)


def bulge(R, Z):
    factor = 2 * (diskprofile(0, 0, 2150.0, 245.0) + diskprofile(0, 0, 3261.0, 743.0))
    distance = (R**2 + Z**2) ** 0.5
    expfunc = np.exp(-distance / 800)
    return factor * expfunc


def halo(R, Z):
    q_h = 0.64
    n_h = 2.77
    f_h = 0.001
    part1 = R**2.0 + (Z / q_h) ** 2.0
    part2 = rsun / np.power(part1, 0.5)
    part3 = np.power(part2, n_h)
    tot = density_rsun * f_h * part3
    return tot


def stellardensity(R, Z, rho=0):
    part1 = thindisk(R, Z)
    part2 = thickdisk(R, Z)
    tot_density = part1 / 1.1 + f / 1.1 * part2 + halo(R, Z) + bulge(R, Z)
    return tot_density


# Juric et al 2008
# Jackson et al 2002



__all__ = ("SNSNRMetric",)

from collections.abc import Iterable

import matplotlib.pylab as plt
import numpy as np
import numpy.lib.recfunctions as rf
from scipy import interpolate

import rubin_sim.maf.metrics as metrics
from rubin_sim.maf.utils.sn_utils import GenerateFakeObservations


class SNSNRMetric(metrics.BaseMetric):
    """
    Metric to estimate the detection rate for faint supernovae
    (x1,color) = (-2.0,0.2)

    Parameters
    ----------
    coadd :  `bool`, optional
        to make "coaddition" per night (uses snStacker)
    lim_sn : class, optional
        Reference data used to simulate LC points (interpolation)
    names_ref : `str`, optional
        names of the simulator used to produce reference data
    season : `float`, optional
        season num
    z : `float`, optional
        redshift for this study
    """

    def __init__(
        self,
        metric_name="SNSNRMetric",
        mjd_col="observationStartMJD",
        ra_col="fieldRA",
        dec_col="fieldDec",
        filter_col="filter",
        m5_col="fiveSigmaDepth",
        exptime_col="visitExposureTime",
        night_col="night",
        obsid_col="observationId",
        nexp_col="numExposures",
        vistime_col="visitTime",
        coadd=True,
        lim_sn=None,
        names_ref=None,
        season=1,
        z=0.01,
        **kwargs,
    ):
        self.mjd_col = mjd_col
        self.m5_col = m5_col
        self.filter_col = filter_col
        self.ra_col = ra_col
        self.dec_col = dec_col
        self.exptime_col = exptime_col
        self.season_col = "season"
        self.night_col = night_col
        self.obsid_col = obsid_col
        self.nexp_col = nexp_col
        self.vistime_col = vistime_col

        cols = [
            self.night_col,
            self.m5_col,
            self.filter_col,
            self.mjd_col,
            self.obsid_col,
            self.nexp_col,
            self.vistime_col,
            self.exptime_col,
            self.season_col,
        ]
        if coadd:
            cols += ["coadd"]
        super(SNSNRMetric, self).__init__(col=cols, metric_name=metric_name, **kwargs)

        self.filter_names = np.array(["u", "g", "r", "i", "z", "y"])
        self.blue_cutoff = 300.0
        self.red_cutoff = 800.0
        self.min_rf_phase = -20.0
        self.max_rf_phase = 40.0
        self.z = z
        self.names_ref = names_ref
        self.season = season

        # SN DayMax: current date - shift days
        self.shift = 10.0

        # These are reference LC
        self.lim_sn = lim_sn

        self.display = False

    def run(self, data_slice, slice_point=None):
        """
        run the metric

        Parameters
        ----------
        data_slice : `np.ndarray`, (N,)
          simulation data under study

        Returns
        -------
        detection rate : `float`
        """
        good_filters = np.isin(data_slice["filter"], self.filter_names)
        data_slice = data_slice[good_filters]
        if data_slice.size == 0:
            return None
        data_slice.sort(order=self.mjd_col)

        if self.season != -1:
            seasons = self.season
        else:
            seasons = np.unique(data_slice["season"])

        if not isinstance(seasons, Iterable):
            seasons = [seasons]

        self.info_season = None
        for seas in seasons:
            info = self.season_info(data_slice, seas)
            if info is not None and info["season_length"] >= self.shift:
                if self.info_season is None:
                    self.info_season = info
                else:
                    self.info_season = np.concatenate((self.info_season, info))

        self.info_season = self.check_seasons(self.info_season)
        if self.info_season is None:
            return 0.0

        sel = data_slice[np.isin(data_slice["season"], np.array(seasons))]

        detect_frac = None
        if len(sel) >= 5:
            detect_frac = self.process(sel)

        if detect_frac is not None:
            return np.median(detect_frac["frac_obs_{}".format(self.names_ref[0])])
        else:
            return 0.0

    def process(self, sel):
        """Process one season

        Parameters
        -----------
        sel : `np.narray`
          array of observations
        season : `int`
          season number

        Returns
        --------
        record array with the following fields:
        fieldRA (float)
        fieldDec (float)
        season (float)
        band (str)
        frac_obs_name_ref (float)
        """

        self.band = np.unique(sel[self.filter_col])[0]
        snr_obs = self.snr_slice(sel)  # SNR for observations
        snr_fakes = self.snr_fakes(sel)  # SNR for fakes
        detect_frac = self.detection_rate(snr_obs, snr_fakes)  # Detection rate
        # snr_obs = np.asarray(snr_obs)
        # snr_fakes = np.asarray(snr_fakes)
        # self.plot(snr_obs, snr_fakes)
        # plt.show()
        detect_frac = np.asarray(detect_frac)

        return detect_frac

    def snr_slice(self, data_slice, j=-1, output_q=None):
        """Estimate SNR for a given data_slice

        Parameters
        -----------
        data_slice : `np.ndarray`
        j : `int`, optional
        output_q : `int`, optional

        Returns
        --------
        snr : `np.ndarray` containing
        SNR_name_ref:  Signal-To-Noise Ratio estimator
        season : season
        cadence: cadence of the season
        season_length: length of the season
        MJD_min: min MJD of the season
        DayMax: SN max luminosity MJD (aka T0)
        MJD:
        m5_eff: mean m5 of obs passing the min_phase, max_phase cut
        field_ra: mean field RA
        field_dec: mean field Dec
        band:  band
        m5: mean m5 (over the season)
        nvisits: median number of visits (per observation) (over the season)
        ExposureTime: median exposure time (per observation) (over the season)
        """

        # Get few infos: RA, Dec, nvisits, m5, exptime
        field_ra = np.mean(data_slice[self.ra_col])
        field_dec = np.mean(data_slice[self.dec_col])
        # one visit = 2 exposures
        nvisits = np.median(data_slice[self.nexp_col] / 2.0)
        m5 = np.mean(data_slice[self.m5_col])
        exptime = np.median(data_slice[self.exptime_col])
        data_slice.sort(order=self.mjd_col)
        mjds = data_slice[self.mjd_col]
        band = np.unique(data_slice[self.filter_col])[0]

        # Define MJDs to consider for metric estimation
        # basically: step of one day between MJDmin and MJDmax
        dates = None

        for val in self.info_season:
            if dates is None:
                dates = np.arange(val["MJD_min"] + self.shift, val["MJD_max"] + 1.0, 1.0)
            else:
                dates = np.concatenate(
                    (
                        dates,
                        np.arange(val["MJD_min"] + self.shift, val["MJD_max"] + 1.0, 1.0),
                    )
                )

        # SN  DayMax: dates-shift where shift is chosen in the input yaml file
        t0_lc = dates - self.shift

        # for these DayMax, estimate the phases of LC points
        # corresponding to the current data_slice MJDs

        time_for_lc = -t0_lc[:, None] + mjds

        phase = time_for_lc / (1.0 + self.z)  # phases of LC points
        # flag: select LC points only in between min_rf_phase and max_phase
        phase_max = self.shift / (1.0 + self.z)
        flag = (phase >= self.min_rf_phase) & (phase <= phase_max)

        # tile m5, MJDs, and seasons to estimate all fluxes and SNR at once
        m5_vals = np.tile(data_slice[self.m5_col], (len(time_for_lc), 1))

        # estimate fluxes and snr in SNR function
        fluxes_tot, snr = self.snr(time_for_lc, m5_vals, flag, t0_lc)

        # now save the results in a record array
        _, idx = np.unique(snr["season"], return_inverse=True)
        infos = self.info_season[idx]

        vars_info = ["cadence", "season_length", "MJD_min"]
        snr = rf.append_fields(snr, vars_info, [infos[name] for name in vars_info])
        snr = rf.append_fields(snr, "DayMax", t0_lc)
        snr = rf.append_fields(snr, "MJD", dates)
        snr = rf.append_fields(snr, "m5_eff", np.mean(np.ma.array(m5_vals, mask=~flag), axis=1))
        global_info = [(field_ra, field_dec, band, m5, nvisits, exptime)] * len(snr)
        names = ["field_ra", "field_dec", "band", "m5", "nvisits", "ExposureTime"]
        global_info = np.rec.fromrecords(global_info, names=names)
        snr = rf.append_fields(snr, names, [global_info[name] for name in names])

        if output_q is not None:
            output_q.put({j: snr})
        else:
            return snr

    def season_info(self, data_slice, season):
        """Get info on seasons for each data_slice

        Parameters
        ----------
        data_slice : `np.ndarray`, (N,)
            array of observations

        Returns
        -------
        info_season : `np.ndarray`
        season, cadence, season_length, MJDmin, MJDmax
        """

        rv = []

        idx = data_slice[self.season_col] == season
        slice_sel = data_slice[idx]
        if len(slice_sel) < 5:
            return None
        slice_sel.sort(order=self.mjd_col)
        mjds_season = slice_sel[self.mjd_col]
        cadence = np.mean(mjds_season[1:] - mjds_season[:-1])
        mjd_min = np.min(mjds_season)
        mjd_max = np.max(mjds_season)
        season_length = mjd_max - mjd_min
        nvisits = np.median(slice_sel[self.nexp_col])
        m5 = np.median(slice_sel[self.m5_col])
        rv.append((float(season), cadence, season_length, mjd_min, mjd_max, nvisits, m5))

        info_season = np.rec.fromrecords(
            rv,
            names=[
                "season",
                "cadence",
                "season_length",
                "MJD_min",
                "MJD_max",
                "nvisits",
                "m5",
            ],
        )

        return info_season

    def snr(self, time_lc, m5_vals, flag, t0_lc):
        """Estimate SNR vs time

        Parameters
        -----------
        time_lc : `np.ndarray`, (N,)
        m5_vals : `list` [`float`]
            five-sigme depth values
        flag : `np.ndarray`, (N,)
            flag to be applied (example: selection from phase cut)
        season_vals : `np.ndarray`, (N,)
            season values
        t0_lc : `np.ndarray`, (N,)
            array of T0 for supernovae

        Returns
        -------
        fluxes_tot : `list` [`float`]
            list of (interpolated) fluxes
        snr_tab : `np.ndarray`, (N,)
            snr_name_ref (float) : Signal-to-Noise values
            season (float) : season num.
        """
        fluxes_tot = {}
        snr_tab = None

        for ib, name in enumerate(self.names_ref):
            fluxes = self.lim_sn.fluxes[ib](time_lc)
            if name not in fluxes_tot.keys():
                fluxes_tot[name] = fluxes
            else:
                fluxes_tot[name] = np.concatenate((fluxes_tot[name], fluxes))

            flux_5sigma = self.lim_sn.mag_to_flux[ib](m5_vals)
            snr = fluxes**2 / flux_5sigma**2
            snr_season = 5.0 * np.sqrt(np.sum(snr * flag, axis=1))

            if snr_tab is None:
                snr_tab = np.asarray(np.copy(snr_season), dtype=[("SNR_" + name, "f8")])
            else:
                snr_tab = rf.append_fields(snr_tab, "SNR_" + name, np.copy(snr_season))
            """
            snr_tab = rf.append_fields(
                snr_tab, 'season', np.mean(seasons, axis=1))
            """
            snr_tab = rf.append_fields(snr_tab, "season", self.get_season(t0_lc))

        # check if any masked value remaining
        # this would correspond to case where no obs point has been selected
        # ie no points with phase in [phase_min,phase_max]
        # this happens when internight gaps are large
        # (typically larger than shift)
        idmask = np.where(snr_tab.mask)
        if len(idmask) > 0:
            tofill = np.copy(snr_tab["season"])
            season_recover = self.get_season(t0_lc[np.where(snr_tab.mask)])
            tofill[idmask] = season_recover
            snr_tab = np.ma.filled(snr_tab, fill_value=tofill)

        return fluxes_tot, snr_tab

    def get_season(self, t0):
        """
        Estimate the seasons corresponding to t0 values

        Parameters
        ----------
        t0 : `list` [`float`]
            set of t0 values

        Returns
        --------
        mean_seasons : `list` [`float`]
            list (float) of corresponding seasons
        """

        diff_min = t0[:, None] - self.info_season["MJD_min"]
        diff_max = -t0[:, None] + self.info_season["MJD_max"]
        seasons = np.tile(self.info_season["season"], (len(diff_min), 1))
        flag = (diff_min >= 0) & (diff_max >= 0)
        seasons = np.ma.array(seasons, mask=~flag)

        return np.mean(seasons, axis=1)

    def snr_fakes(self, data_slice):
        """Estimate SNR for fake observations
        in the same way as for observations (using SNR_Season)

        Parameters
        -----------
        data_slice : `np.ndarray`, (N,)
            array of observations

        Returns
        --------
        snr_tab : `np.ndarray`
            snr_name_ref (float) : Signal-to-Noise values
            season (float) : season num.

        """

        # generate fake observations
        fake_obs = None

        # idx = (data_slice[self.season_col] == season)
        band = np.unique(data_slice[self.filter_col])[0]
        fake_obs = self.gen_fakes(data_slice, band)

        # estimate SNR vs MJD

        snr_fakes = self.snr_slice(fake_obs[fake_obs["filter"] == band])

        return snr_fakes

    def gen_fakes(self, slice_sel, band):
        """Generate fake observations
        according to observing values extracted from simulations

        Parameters
        ----------
        slice_sel : `np.ndarray`, (N,)
            array of observations
        band : `str`
            band to consider

        Returns
        --------
        fake_obs_season : `np.ndarray`
            array of observations with the following fields
            observationStartMJD (float)
            field_ra (float)
            field_dec (float)
            filter (U1)
            fiveSigmaDepth (float)
            numExposures (float)
            visitExposureTime (float)
            season (int)
        """
        field_ra = np.mean(slice_sel[self.ra_col])
        field_dec = np.mean(slice_sel[self.dec_col])
        tvisit = 30.0

        fake_obs = None
        for val in self.info_season:
            cadence = val["cadence"]
            mjd_min = val["MJD_min"]
            season_length = val["season_length"]
            nvisits = val["nvisits"]
            m5 = val["m5"]

            # build the configuration file

            config_fake = {}
            config_fake["Ra"] = field_ra
            config_fake["Dec"] = field_dec
            config_fake["bands"] = [band]
            config_fake["Cadence"] = [cadence]
            config_fake["MJD_min"] = [mjd_min]
            config_fake["season_length"] = season_length
            config_fake["nvisits"] = [nvisits]
            m5_nocoadd = m5 - 1.25 * np.log10(float(nvisits) * tvisit / 30.0)
            config_fake["m5"] = [m5_nocoadd]
            config_fake["seasons"] = [val["season"]]
            config_fake["Exposure_Time"] = [30.0]
            config_fake["shift_days"] = 0.0
            fake_obs_season = GenerateFakeObservations(config_fake).Observations
            if fake_obs is None:
                fake_obs = fake_obs_season
            else:
                fake_obs = np.concatenate((fake_obs, fake_obs_season))
        return fake_obs

    def plot(self, snr_obs, snr_fakes):
        """Plot SNR vs time

        Parameters
        ----------
        snr_obs : `np.ndarray`, (N,)
            array estimated using snr_slice(observations)
        snr_obs : `np.ndarray`, (N,)
            array estimated using snr_slice(fakes)
        """

        fig, ax = plt.subplots(figsize=(10, 7))

        title = "season {} - {} band - z={}".format(self.season, self.band, self.z)
        fig.suptitle(title)
        ax.plot(
            snr_obs["MJD"],
            snr_obs["SNR_{}".format(self.names_ref[0])],
            label="Simulation",
        )
        ax.plot(
            snr_fakes["MJD"],
            snr_fakes["SNR_{}".format(self.names_ref[0])],
            ls="--",
            label="Fakes",
        )

    def plot_history(self, fluxes, mjd, flag, snr, t0_lc, dates):
        """Plot history of lightcurve
        For each MJD, fluxes and snr are plotted
        Each plot may be saved as a png to make a video afterwards

        Parameters
        ----------
        fluxes : `list` [`float`]
            LC fluxes
        mjd : list(float)
            mjds of the fluxes
        flag : array
            flag for selection of fluxes
        snr : list
            signal-to-noise ratio
        t0_lc : list(float)
            list of T0 supernovae
        dates : list(float)
            date of the display (mjd)
        """

        import matplotlib.pyplot as plt

        plt.ion()
        fig, ax = plt.subplots(ncols=1, nrows=2)
        fig.canvas.draw()

        colors = ["b", "r"]
        myls = ["-", "--"]
        tot_label = []
        fontsize = 12
        mjd_ma = np.ma.array(mjd, mask=~flag)
        fluxes_ma = {}
        for key, val in fluxes.items():
            fluxes_ma[key] = np.ma.array(val, mask=~flag)
        key = list(fluxes.keys())[0]
        jmax = len(fluxes_ma[key])
        tot_label = []
        tot_label_snr = []
        min_flux = []
        max_flux = []

        for j in range(jmax):
            for ib, name in enumerate(fluxes_ma.keys()):
                tot_label.append(
                    ax[0].errorbar(
                        mjd_ma[j],
                        fluxes_ma[name][j],
                        marker="s",
                        color=colors[ib],
                        ls=myls[ib],
                        label=name,
                    )
                )

                tot_label_snr.append(
                    ax[1].errorbar(
                        snr["MJD"][:j],
                        snr["SNR_" + name][:j],
                        color=colors[ib],
                        label=name,
                    )
                )
                fluxx = fluxes_ma[name][j]
                fluxx = fluxx[~fluxx.mask]
                if len(fluxx) >= 2:
                    min_flux.append(np.min(fluxx))
                    max_flux.append(np.max(fluxx))
                else:
                    min_flux.append(0.0)
                    max_flux.append(200.0)

            min_fluxes = np.min(min_flux)
            max_fluxes = np.max(max_flux)

            tot_label.append(
                ax[0].errorbar(
                    [t0_lc[j], t0_lc[j]],
                    [min_fluxes, max_fluxes],
                    color="k",
                    label="DayMax",
                )
            )
            tot_label.append(
                ax[0].errorbar(
                    [dates[j], dates[j]],
                    [min_fluxes, max_fluxes],
                    color="k",
                    ls="--",
                    label="Current MJD",
                )
            )
            fig.canvas.flush_events()
            # plt.savefig('{}/{}_{}.png'.format(dir_save, 'snr', 1000 + j))
            if j != jmax - 1:
                ax[0].clear()
                tot_label = []
                tot_label_snr = []

        labs = [ll.get_label() for ll in tot_label]
        ax[0].legend(tot_label, labs, ncol=1, loc="best", prop={"size": fontsize}, frameon=False)
        ax[0].set_ylabel("Flux [e.sec$^{-1}$]", fontsize=fontsize)

        ax[1].set_xlabel("MJD", fontsize=fontsize)
        ax[1].set_ylabel("SNR", fontsize=fontsize)
        ax[1].legend()
        labs = [ll.get_label() for ll in tot_label_snr]
        ax[1].legend(
            tot_label_snr,
            labs,
            ncol=1,
            loc="best",
            prop={"size": fontsize},
            frameon=False,
        )
        for i in range(2):
            ax[i].tick_params(axis="x", labelsize=fontsize)
            ax[i].tick_params(axis="y", labelsize=fontsize)

    def detection_rate(self, snr_obs, snr_fakes):
        """
        Estimate the time fraction(per season) for which
        snr_obs > snr_fakes = detection rate
        For regular cadences one should get a result close to 1

        Parameters
        ----------
        snr_obs : array
            array estimated using snr_slice(observations)

        snr_fakes: array
            array estimated using snr_slice(fakes)

        Returns
        -------
        record array with the following fields:
            fieldRA (float)
            fieldDec (float)
            season (float)
            band (str)
            frac_obs_name_ref (float)
        """

        ra = np.mean(snr_obs["fieldRA"])
        dec = np.mean(snr_obs["fieldDec"])
        band = np.unique(snr_obs["band"])[0]

        rtot = []

        for season in np.unique(snr_obs["season"]):
            idx = snr_obs["season"] == season
            sel_obs = snr_obs[idx]
            idxb = snr_fakes["season"] == season
            sel_fakes = snr_fakes[idxb]

            sel_obs.sort(order="MJD")
            sel_fakes.sort(order="MJD")
            r = [ra, dec, season, band]
            names = [self.ra_col, self.dec_col, "season", "band"]
            for sim in self.names_ref:
                fakes = interpolate.interp1d(sel_fakes["MJD"], sel_fakes["SNR_" + sim])
                obs = interpolate.interp1d(sel_obs["MJD"], sel_obs["SNR_" + sim])
                mjd_min = np.max([np.min(sel_obs["MJD"]), np.min(sel_fakes["MJD"])])
                mjd_max = np.min([np.max(sel_obs["MJD"]), np.max(sel_fakes["MJD"])])
                mjd = np.arange(mjd_min, mjd_max, 1.0)

                diff_res = obs(mjd) - fakes(mjd)

                idx = diff_res >= 0
                r += [len(diff_res[idx]) / len(diff_res)]
                names += ["frac_obs_" + sim]
            rtot.append(tuple(r))

        return np.rec.fromrecords(rtot, names=names)

    def check_seasons(self, tab):
        """Check whether seasons have no overlap
        if it is the case: modify MJD_min and season length of
        the corresponding season
        return only seasons with season_length > 30 days

        Parameters
        -----------
        tab : array with the following fields:

        Returns
        -------
        tab : array with the following fields:
        """
        if tab is None or len(tab) == 1:
            return tab

        if len(tab) > 1:
            diff = tab["MJD_min"][1:] - tab["MJD_max"][:-1]
            idb = np.argwhere(diff < 20.0)
            if len(idb) >= 1:
                tab["MJD_min"][idb + 1] = tab["MJD_max"][idb] + 20.0
                tab["season_length"][idb + 1] = tab["MJD_max"][idb + 1] - tab["MJD_min"][idb + 1]

            return tab[tab["season_length"] > 30.0]



"""Metrics for scheduler monitoring and progress."""

__all__ = ["AgeMetric"]

import numpy as np

from .base_metric import BaseMetric


class AgeMetric(BaseMetric):
    def __init__(
        self, mjd, mjd_col="observationStartMJD", long_limit=30, metric_name="age", mask_val=np.nan, **kwargs
    ):
        """Metric that shows the time since the previous visit in each slice,
        as of a given time

        Parameters
        ----------
        mjd : `float`
            Reference time for the age.
        mjd_col : `str`
            Column with the time of visit, by default "observationStartMJD"
        long_limit : `int`
            The age past which to mask values, by default 30
        metric_name : `str`
            The metric name, by default 'age'
        mask_val : `object`
            Name for masked values, by default np.nan
        """
        self.mjd = mjd
        self.mjd_col = mjd_col
        self.long_limit = long_limit
        super().__init__(col=[self.mjd_col], metric_name=metric_name, mask_val=mask_val, **kwargs)

    def run(self, data_slice, slice_point=None):
        age = self.mjd - np.max(data_slice[self.mjd_col])
        if age > self.long_limit:
            age = self.mask_val
        return age



__all__ = ("ra_dec2_alt_az", "GalacticStacker", "EclipticStacker")

import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord, get_sun
from astropy.time import Time
from rubin_scheduler.utils import calc_lmst

from .base_stacker import BaseStacker


def wrap_ra(ra):
    """
    Wrap only RA values into 0-2pi (using mod).

    Parameters
    ----------
    ra : numpy.ndarray
        RA in radians

    Returns
    -------
    numpy.ndarray
        Wrapped RA values, in radians.
    """
    ra = ra % (2.0 * np.pi)
    return ra


def ra_dec2_alt_az(ra, dec, lat, lon, mjd, altonly=False):
    """Convert RA/Dec (and telescope site lat/lon) to alt/az.

    This uses simple equations and ignores aberation, precession, nutation.

    Parameters
    ----------
    ra : `np.ndarray`, (N,)
        RA, in radians.
    dec : `np.ndarray`, (N,)
        Dec, in radians. Must be same length as `ra`.
    lat : `float`
        Latitude of the observatory in radians.
    lon : `float`
        Longitude of the observatory in radians.
    mjd : `float`
        Modified Julian Date.
    altonly : `bool`, optional
        Calculate altitude only.

    Returns
    -------
    alt : `np.ndarray`, (N,)
        Altitude, same length as `ra` and `dec`. Radians.
    az : `np.ndarray`, (N,)
        Azimuth, same length as `ra` and `dec`. Radians.
    """
    lmst = calc_lmst(mjd, lon)
    lmst = lmst / 12.0 * np.pi  # convert to rad
    ha = lmst - ra
    sindec = np.sin(dec)
    sinlat = np.sin(lat)
    coslat = np.cos(lat)
    sinalt = sindec * sinlat + np.cos(dec) * coslat * np.cos(ha)
    # make sure sinalt is in the expected range.
    sinalt = np.where(sinalt < -1, -1, sinalt)
    sinalt = np.where(sinalt > 1, 1, sinalt)
    alt = np.arcsin(sinalt)
    if altonly:
        az = None
    else:
        cosaz = (sindec - np.sin(alt) * sinlat) / (np.cos(alt) * coslat)
        cosaz = np.where(cosaz < -1, -1, cosaz)
        cosaz = np.where(cosaz > 1, 1, cosaz)
        az = np.arccos(cosaz)
        signflip = np.where(np.sin(ha) > 0)
        az[signflip] = 2.0 * np.pi - az[signflip]
    return alt, az


class GalacticStacker(BaseStacker):
    """Add the galactic coordinates of each RA/Dec pointing: gall, galb

    Parameters
    ----------
    ra_col : str, optional
        Name of the RA column. Default fieldRA.
    dec_col : str, optional
        Name of the Dec column. Default fieldDec.
    """

    cols_added = ["gall", "galb"]

    def __init__(self, ra_col="fieldRA", dec_col="fieldDec", degrees=True):
        self.cols_req = [ra_col, dec_col]
        self.ra_col = ra_col
        self.dec_col = dec_col
        self.degrees = degrees
        if self.degrees:
            self.units = ["degrees", "degrees"]
        else:
            self.units = ["radians", "radians"]

    def _run(self, sim_data, cols_present=False):
        # raCol and DecCol in radians, gall/b in radians.
        if cols_present:
            # Column already present in data;
            # assume it is correct and does not need recalculating.
            return sim_data
        if self.degrees:
            c = SkyCoord(ra=sim_data[self.ra_col] * u.deg, dec=sim_data[self.dec_col] * u.deg).transform_to(
                "galactic"
            )
        else:
            c = SkyCoord(ra=sim_data[self.ra_col] * u.rad, dec=sim_data[self.dec_col] * u.rad).transform_to(
                "galactic"
            )
        sim_data["gall"] = c.l.rad
        sim_data["galb"] = c.b.rad
        return sim_data


class EclipticStacker(BaseStacker):
    """Add the ecliptic coordinates of each RA/Dec pointing:
    eclipLat, eclipLon

    Optionally subtract off the sun's ecliptic longitude and wrap.

    Parameters
    ----------
    mjd_col : str, optional
        Name of the MJD column. Default expMJD.
    ra_col : str, optional
        Name of the RA column. Default fieldRA.
    dec_col : str, optional
        Name of the Dec column. Default fieldDec.
    subtract_sun_lon : bool, optional
        Flag to subtract the sun's ecliptic longitude. Default False.
    """

    cols_added = ["eclipLat", "eclipLon"]

    def __init__(
        self,
        mjd_col="observationStartMJD",
        ra_col="fieldRA",
        dec_col="fieldDec",
        degrees=True,
        subtract_sun_lon=False,
    ):
        self.cols_req = [mjd_col, ra_col, dec_col]
        self.subtract_sun_lon = subtract_sun_lon
        self.degrees = degrees
        if self.degrees:
            self.units = ["degrees", "degrees"]
        else:
            self.units = ["radians", "radians"]
        self.mjd_col = mjd_col
        self.ra_col = ra_col
        self.dec_col = dec_col

    def _run(self, sim_data, cols_present=False):
        if cols_present:
            # Column already present in data;
            # assume it is correct and does not need recalculating.
            return sim_data
        for i in np.arange(sim_data.size):
            if self.degrees:
                coord = SkyCoord(
                    ra=sim_data[self.ra_col] * u.degree,
                    dec=sim_data[self.dec_col] * u.degree,
                )
            else:
                coord = SkyCoord(ra=sim_data[self.ra_col] * u.rad, dec=sim_data[self.dec_col] * u.rad)
            coord_ecl = coord.geocentricmeanecliptic
            sim_data["eclipLat"] = coord_ecl.lat.rad

            if self.subtract_sun_lon:
                times = Time(sim_data[self.mjd_col])
                sun = get_sun(times)
                sun_ecl = sun.geocentricmeanecliptic
                lon = wrap_ra(coord_ecl.lon.rad - sun_ecl.lon.rad)
                sim_data["eclipLon"] = lon
            else:
                sim_data["eclipLon"] = coord_ecl.lon.rad
        if self.degrees:
            sim_data["eclipLon"] = np.degrees(sim_data["eclipLon"])
            sim_data["eclipLat"] = np.degrees(sim_data["eclipLat"])
        return sim_data



__all__ = ("stellar_mags",)

import numpy as np


def calc_wd_colors():
    """
    Calculate a few example WD colors. Values to go in stellar_mags().
    Here in case values need to be regenerated
    (different stars, bandpasses change, etc.)
    """

    try:
        import os

        from lsst.utils import getPackageDir

        from rubin_sim.photUtils import Bandpass, Sed
    except ImportError:
        "Need to setup sims_photUtils to generate WD magnitudes."

    names = ["HeWD_25200_80", "WD_11000_85", "WD_3000_85"]
    fns = [
        "bergeron_He_24000_80.dat_25200.gz",
        "bergeron_10500_85.dat_11000.gz",
        "bergeron_2750_85.dat_3000.gz",
    ]
    wd_dir = os.path.join(getPackageDir("sims_sed_library"), "starSED/wDs/")
    files = [os.path.join(wd_dir, filename) for filename in fns]

    # Read in the LSST bandpasses
    bp_names = ["u", "g", "r", "i", "z", "y"]
    bps = []
    through_path = os.path.join(getPackageDir("throughputs"), "baseline")
    for key in bp_names:
        bp = np.loadtxt(
            os.path.join(through_path, "filter_" + key + ".dat"),
            dtype=list(zip(["wave", "trans"], [float] * 2)),
        )
        temp_b = Bandpass()
        temp_b.setBandpass(bp["wave"], bp["trans"])
        bps.append(temp_b)

    # Read in the SEDs and compute mags
    mags = []
    for filename in files:
        star = Sed()
        star.readSED_flambda(filename)
        single_mags = [star.calcMag(band) for band in bps]
        mags.append([single_mags[i - 1] - single_mags[i] for i in range(1, 6)])

    for maglist, fn, name in zip(mags, fns, names):
        format = (name, fn) + tuple(maglist)
        print("['%s', '%s', %f, %f, %f, %f, %f]" % format)


def stellar_mags(stellar_type, rmag=19.0):
    """
    Calculates the expected magnitudes in LSST filters for a
    typical star of the given spectral type.

    Based on mapping of Kuruz models to spectral types here:
    http://www.stsci.edu/hst/observatory/crds/k93models.html


    Parameters
    ----------
    stellar_type : str
        Spectral type of a star (O,B,A,F,G,K,M), or for white dwarf colors,
        one of 'HeWD_25200_80, 'WD_11000_85', 'WD_3000_85'
    rmag : float
        The expected r-band magnitude of the star.

    Returns
    -------
    dict of floats
        The expected magnitudes in LSST filters.
    """

    # If this is the first time running the function, set up the data array
    if not hasattr(stellar_mags, "data"):
        names = ["stellar_type", "Model Name", "u-g", "g-r", "r-i", "i-z", "z-y"]
        types = [("U", 20), ("U", 35), float, float, float, float, float]
        data = np.core.records.fromrecords(
            [
                (
                    "O",
                    "kp00_50000[g50]",
                    -0.4835688497,
                    -0.5201721327,
                    -0.3991733698,
                    -0.3106800468,
                    -0.2072290744,
                ),
                (
                    "B",
                    "kp00_30000[g40]",
                    -0.3457202828,
                    -0.4834762052,
                    -0.3812792176,
                    -0.2906072887,
                    -0.1927230035,
                ),
                (
                    "A",
                    "kp00_9500[g40]",
                    0.8823182684,
                    -0.237288029,
                    -0.2280783991,
                    -0.1587960264,
                    -0.03043824335,
                ),
                (
                    "F",
                    "kp00_7250[g45]",
                    0.9140316091,
                    0.1254277486,
                    -0.03419150003,
                    -0.0802010739,
                    -0.03802756413,
                ),
                (
                    "G",
                    "kp00_6000[g45]",
                    1.198219095,
                    0.3915608688,
                    0.09129426676,
                    0.002604263747,
                    -0.004659443668,
                ),
                (
                    "K",
                    "kp00_5250[g45]",
                    1.716635024,
                    0.6081567546,
                    0.1796910856,
                    0.06492278686,
                    0.0425155827,
                ),
                (
                    "M",
                    "kp00_3750[g45]",
                    2.747842719,
                    1.287599638,
                    0.5375622482,
                    0.4313486709,
                    0.219308065,
                ),
                (
                    "HeWD_25200_80",
                    "bergeron_He_24000_80.dat_25200.gz",
                    -0.218959,
                    -0.388374,
                    -0.326946,
                    -0.253573,
                    -0.239460,
                ),
                (
                    "WD_11000_85",
                    "bergeron_10500_85.dat_11000.gz",
                    0.286146,
                    -0.109115,
                    -0.178500,
                    -0.185833,
                    -0.186913,
                ),
                (
                    "WD_3000_85",
                    "bergeron_2750_85.dat_3000.gz",
                    3.170620,
                    1.400062,
                    0.167195,
                    0.127024,
                    -0.378069,
                ),
            ],
            dtype=list(zip(names, types)),
        )
        # Switch to a dict for faster look-up
        stellar_mags.data = {}
        for row in data:
            stellar_mags.data["%s" % row["stellar_type"]] = row

    results = {}
    # good = np.where(stellar_mags.data['stellar_type'] == stellar_type)
    if stellar_type not in stellar_mags.data:
        message = "Received stellar_type %s" % stellar_type
        message += " but expected one of %s" % ", ".join(stellar_mags.data.keys())
        raise ValueError(message)

    results["r"] = rmag
    results["i"] = rmag - stellar_mags.data[stellar_type]["r-i"]
    results["z"] = results["i"] - stellar_mags.data[stellar_type]["i-z"]
    results["y"] = results["z"] - stellar_mags.data[stellar_type]["z-y"]
    results["g"] = stellar_mags.data[stellar_type]["g-r"] + results["r"]
    results["u"] = stellar_mags.data[stellar_type]["u-g"] + results["g"]
    return results



import abc
import warnings
from collections import defaultdict
from numbers import Integral
from weakref import WeakKeyDictionary

import astropy.units as u
import healpy as hp
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import numpy.ma as ma
import pandas as pd
import skyproj
from astropy.coordinates import SkyCoord
from rubin_scheduler.utils import _healbin

from .plot_handler import BasePlotter, apply_zp_norm
from .spatial_plotters import set_color_lims, set_color_map


def compute_circle_points(
    center_ra,
    center_decl,
    radius=90.0,
    start_bear=0,
    end_bear=360,
    step=1,
):
    """Create points along a circle or arc on a sphere

    Parameters
    ----------
    center_ra : `float`
        R.A. of the center of the circle (deg.).
    center_decl : `float`
        Decl. of the center of the circle (deg.).
    radius : float, optional
        Radius of the circle (deg.), by default 90.0
    start_bear : int, optional
        Bearing (E. of N.) of the start of the circle (deg.), by default 0
    end_bear : int, optional
        Bearing (E. of N.) of the end of the circle (deg.), by default 360
    step : int, optional
        Spacing of the points along the circle (deg.), by default 1

    Returns
    -------
    circle : `pandas.DataFrame`
        DataFrame with points in the circle.
    """
    ras = []
    decls = []

    bearing_angles = np.arange(start_bear, end_bear + step, step) * u.deg
    center_coords = SkyCoord(center_ra * u.deg, center_decl * u.deg)
    circle_coords = center_coords.directional_offset_by(bearing_angles, radius * u.deg)
    bearings = bearing_angles.value.tolist()
    ras = circle_coords.ra.deg.tolist()
    decls = circle_coords.dec.deg.tolist()

    # de-normalize RA so there are no large jumps, which can
    # confuse cartopy + matplotlib

    previous_ra = ras[0]
    for ra_index in range(1, len(ras)):
        if ras[ra_index] - previous_ra > 180:
            ras[ra_index] -= 360
        if ras[ra_index] - previous_ra < -180:
            ras[ra_index] += 360
        previous_ra = ras[ra_index]

    circle = pd.DataFrame(
        {
            "bearing": bearings,
            "ra": ras,
            "decl": decls,
        }
    ).set_index("bearing")

    return circle


class SkyprojPlotter(BasePlotter, abc.ABC):
    """Base class for all plotters that use the skyproj module."""

    # Store a mapping between figures and existing SkyprojPlotters, but use
    # weak keys so that the figure can be cleaned up by the garbage collector,
    # and when it is entry in the dictionary automatically deleted.
    skyproj_instances = WeakKeyDictionary()

    default_decorations = ["ecliptic", "galactic_plane"]

    default_colorbar_kwargs = {
        "location": "bottom",
        "shrink": 0.75,
        "aspect": 25,
        "pad": 0.1,
        "orientation": "horizontal",
    }

    ecliptic_kwargs = {"edgecolor": "green"}
    galactic_plane_kwargs = {"edgecolor": "blue"}
    sun_kwargs = {"color": "yellow"}
    moon_kwargs = {"color": "orange"}
    horizon_kwargs = {"edgecolor": "black", "linewidth": 3}

    def __init__(self):
        super().__init__()
        # Customize our plotters members for our new plot

        self.plot_type = "Skyproj"
        self.default_plot_dict = {}
        self.default_plot_dict.update(
            {
                "subplot": 111,
                "skyproj": skyproj.MollweideSkyproj,
                "skyproj_kwargs": {"lon_0": 0},
                "decorations": self.default_decorations,
                "title": None,
                "xlabel": None,
                "label": None,
                "labelsize": None,
                "fontsize": None,
                "figsize": None,
            }
        )
        self._initialize_plot_dict({})
        self.figure = None

    def _initialize_plot_dict(self, user_plot_dict):
        # Use a defaultdict so that everything not explicitly set return None.
        def return_none():
            return None

        self.plot_dict = defaultdict(return_none)
        self.plot_dict.update(self.default_plot_dict)
        self.plot_dict.update(user_plot_dict)

    def _prepare_skyproj(self, fig=None):
        # Exctract elemens of plot_dict that need to be passed to plt.figure
        fig_kwargs = {k: v for k, v in self.plot_dict.items() if k in ["figsize"]}

        self.figure = plt.figure(**fig_kwargs) if fig is None else fig

        # Normalize subplot specification,
        # so that it can be set in either (single) integer form or tuple form.
        # This is needed so that the same subplot set by different ways will be
        # stored in the same key in the instance dictionary.
        subplot_in = self.plot_dict["subplot"]
        in_integral_form = isinstance(subplot_in, Integral) and len(str(subplot_in)) == 3
        subplot = tuple(map(int, str(subplot_in))) if in_integral_form else tuple(subplot_in)

        # If there is an existing instance of SkyprojPlotter (or one of its
        # subclasses), reuse it.
        # Otherwise, create a new one and store it so that it may be reused.
        if self.figure not in self.skyproj_instances:
            self.skyproj_instances[self.figure] = {}

        if subplot in self.skyproj_instances[self.figure]:
            existing_instance = self.skyproj_instances[self.figure][subplot]
            assert isinstance(existing_instance, self.plot_dict["skyproj"])
            self.skyproj = existing_instance
        else:
            # This instance of Axis will get deleted and replaced when the
            # corresponding SkyPlot class is created, but this instances is
            # needed so SkyPlot will put its instance in the correct relation
            # with its figure.
            ax = self.figure.add_subplot(*subplot)
            self.skyproj = self.plot_dict["skyproj"](ax=ax, **self.plot_dict["skyproj_kwargs"])
            self.skyproj_instances[self.figure][subplot] = self.skyproj

    def draw_circle(self, center_ra, center_decl, radius=90, **kwargs):
        """Draw a circle on the sphere.

        Parameters
        ----------
        center_ra : `float`
            R.A. of the center of the circle (deg.).
        center_decl : `float`
            Decl. of the center of the circle (deg.).
        radius : float, optional
            Radius of the circle (deg.), by default 90.0
        **kwargs
            Additional keyword arguments passed to
            `skyproj._Skyproj.draw_polygon`.
        """
        points = compute_circle_points(center_ra, center_decl, radius)
        self.skyproj.draw_polygon(points.ra, points.decl, **kwargs)

    def draw_ecliptic(self):
        """Draw the ecliptic the sphere.

        Parameters
        ----------
        **kwargs
            Keyword arguments passed to
            `skyproj._Skyproj.draw_polygon`.
        """
        ecliptic_pole = SkyCoord(lon=0 * u.degree, lat=90 * u.degree, frame="geocentricmeanecliptic").icrs
        self.draw_circle(ecliptic_pole.ra.deg, ecliptic_pole.dec.deg, **self.ecliptic_kwargs)

    def draw_galactic_plane(self):
        """Draw the galactic plane the sphere.

        Parameters
        ----------
        **kwargs
            Keyword arguments passed to
            `skyproj._Skyproj.draw_polygon`.
        """
        galactic_pole = SkyCoord(l=0 * u.degree, b=90 * u.degree, frame="galactic").icrs
        self.draw_circle(galactic_pole.ra.deg, galactic_pole.dec.deg, **self.galactic_plane_kwargs)

    def draw_zd(self, zd=90, **kwargs):
        """Draw a circle at a given zenith distance.

        Parameters
        ----------
        zd : `float`
            The zenith distance to draw, in degrees.
            Defaults to 90.
        **kwargs
            Keyword arguments passed to
            `skyproj._Skyproj.draw_polygon`.

        Notes
        -----
        This method relies on the site location and time (as an mjd) set in
        the ``model_observatory`` element of the ``plot_dict``, which
        should be of the class
        `rubin_scheduler.scheduler.model_observatory.ModelObservatory`.
        """
        if "model_observatory" not in self.plot_dict or self.plot_dict["model_observatory"] is None:
            warnings.warn("plot_dict['model_observatory'] must be set to plot a zenith distance circle")
            return

        model_observatory = self.plot_dict["model_observatory"]
        lmst = model_observatory.return_conditions().lmst * 360 / 24
        latitude = model_observatory.location.lat.deg
        self.draw_circle(lmst, latitude, zd, **kwargs)

    def draw_body(self, body="sun", **kwargs):
        """Mark the sun or moon.

        Parameters
        ----------
        body : `str`
            The name of the body to draw, either `sun` or `moon`.
            Defaults to `sun`
        **kwargs
            Keyword arguments passed to
            `skyproj._Skyproj.scatter`

        Notes
        -----
        This method relies on the site location and time (as an mjd) set in
        the ``model_observatory`` element of the ``plot_dict``, which
        should be of the class
        `rubin_scheduler.scheduler.model_observatory.ModelObservatory`.
        """
        if "model_observatory" not in self.plot_dict or self.plot_dict["model_observatory"] is None:
            warnings.warn(f"plot_dict['model_observatory'] must be set to plot the {body}")
            return

        mjd = self.plot_dict["model_observatory"].mjd
        model_observatory = self.plot_dict["model_observatory"]
        sun_moon_positions = model_observatory.almanac.get_sun_moon_positions(mjd)
        ra = np.degrees(sun_moon_positions[f"{body}_RA"].item())
        decl = np.degrees(sun_moon_positions[f"{body}_dec"].item())
        self.skyproj.scatter(ra, decl, **kwargs)

    def decorate(self):
        """Add decorations/annotations to the sky plot.

        Notes
        -----
        For decorations that depend on the time or location of the observer
        (e.g., the horizon and sun and moon positions)
        this method relies on the site location and time (as an mjd) set in
        the ``model_observatory`` element of the ``plot_dict``, which
        should be of the class
        `rubin_scheduler.scheduler.model_observatory.ModelObservatory`.
        Other decorations (e.g., the ecliptic and galactic plane) can
        be shown even when ``model_observatory`` is not set.
        """
        decorations = self.plot_dict["decorations"]

        if "ecliptic" in decorations:
            self.draw_ecliptic()

        if "galactic_plane" in decorations:
            self.draw_galactic_plane()

        if "sun" in decorations:
            self.draw_body("sun", **self.sun_kwargs)

        if "moon" in decorations:
            self.draw_body("moon", **self.moon_kwargs)

        if "horizon" in decorations:
            self.draw_zd(**self.horizon_kwargs)

    @abc.abstractmethod
    def draw(self, metric_values, slicer):
        raise NotImplementedError()

    def __call__(self, metric_values, slicer, user_plot_dict, fig=None):
        """Make a plot.

        Parameters
        ----------
        metric_value : `numpy.ma.MaskedArray`
            The metric values from the bundle.
        slicer : `rubin_sim.maf.slicers.TwoDSlicer`
            The slicer.
        user_plot_dict: `dict`
            Dictionary of plot parameters set by user
            (overrides default values).
        fig : `matplotlib.figure.Figure`
            Matplotlib figure.
            The default is ``None``, which starts new figure.

        Returns
        -------
        fig : `matplotlib.figure.Figure`
           Figure with the plot.
        """
        self._initialize_plot_dict(user_plot_dict)
        self._prepare_skyproj(fig)
        self.draw(metric_values, slicer)
        self.decorate()

        # Do not show axis labels unless they are set explicitly
        # or specified as decorators.
        if self.plot_dict["xlabel"] is None:
            if "xlabel" not in self.plot_dict["decorations"]:
                self.skyproj.ax.set_xlabel("", visible=False)
        else:
            self.skyproj.ax.set_xlabel(self.plot_dict["xlabel"])

        if self.plot_dict["ylabel"] is None:
            if "ylabel" not in self.plot_dict["decorations"]:
                self.skyproj.ax.set_ylabel("", visible=False)
        else:
            self.skyproj.ax.set_ylabel(self.plot_dict["ylabel"])

        if self.plot_dict["label"] is not None:
            label_kwargs = {}
            if "fontsize" in self.plot_dict:
                label_kwargs["fontsize"] = self.plot_dict["fontsize"]
            if "label_loc" in self.plot_dict:
                label_kwargs["loc"] = self.plot_dict["label_loc"]
            self.skyproj.ax.set_title(self.plot_dict["label"], **label_kwargs)

        if self.plot_dict["title"] is not None:
            title_kwargs = {}
            if "fontsize" in self.plot_dict:
                title_kwargs["fontsize"] = self.plot_dict["fontsize"]
            self.figure.suptitle(self.plot_dict["title"], **title_kwargs)

        return self.figure


class HpxmapPlotter(SkyprojPlotter):
    default_hpixmap_kwargs = {"zoom": False}

    def __init__(self):
        super().__init__()
        self.default_plot_dict["colorbar"] = True
        self.object_plotter = False

    def draw_colorbar(self):
        """Add a color bar."""
        plot_dict = self.plot_dict

        colorbar_kwargs = {}
        colorbar_kwargs.update(self.default_colorbar_kwargs)

        if "extend" in self.plot_dict:
            colorbar_kwargs["extendrect"] = self.plot_dict["extend"] == "neither"

        if "cbar_format" in self.plot_dict:
            colorbar_kwargs["format"] = self.plot_dict["cbar_format"]

        if "cbar_orientation" in self.plot_dict:
            colorbar_kwargs["orientation"] = self.plot_dict["cbar_orientation"]
            match self.plot_dict["cbar_orientation"].lower():
                case "vertical":
                    colorbar_kwargs["shrink"] = 0.5
                    colorbar_kwargs["location"] = "right"
                case "horizontal":
                    colorbar_kwargs["location"] = "bottom"
                case _:
                    orientation = self.plot_dict["cbar_orientation"]
                    raise NotImplementedError(f"cbar_orintation {orientation} is not supported")

        colorbar = self.skyproj.draw_colorbar(**colorbar_kwargs)

        colorbar.set_label(self.plot_dict["xlabel"], fontsize=self.plot_dict["fontsize"])

        if "labelsize" in self.plot_dict and self.plot_dict["labelsize"] is not None:
            colorbar.ax.tick_params(labelsize=plot_dict["labelsize"])

        if "n_ticks" in self.plot_dict and self.plot_dict["n_ticks"] is not None:
            tick_locator = mpl.ticker.MaxNLocator(nbins=plot_dict["n_ticks"])
            colorbar.locator = tick_locator
            colorbar.update_ticks()

        # If outputing to PDF, this fixes the colorbar white stripes
        if "cbar_edge" in self.plot_dict and self.plot_dict["cbar_edge"]:
            colorbar.solids.set_edgecolor("face")

    def decorate(self):
        super().decorate()

        if self.plot_dict["colorbar"]:
            self.draw_colorbar()

    def _transform_to_healpix(self, metric_value_in, slicer):
        # Bin the values up on a healpix grid.
        metric_value = _healbin(
            slicer.slice_points["ra"],
            slicer.slice_points["dec"],
            metric_value_in.filled(slicer.badval),
            nside=self.plot_dict["nside"],
            reduce_func=self.plot_dict["reduce_func"],
            fill_val=slicer.badval,
        )
        mask = np.zeros(metric_value.size)
        mask[np.where(metric_value == slicer.badval)] = 1

        metric_value = ma.array(metric_value, mask=mask)
        return metric_value

    def _compatible_color_limits(self, norm, clims):
        # Log can't be 0
        compatible = (norm != "log") or (clims[0] < 0 and clims[1] < 0) or (clims[0] > 0 and clims[1] > 0)
        return compatible

    def draw(self, metric_values_in, slicer):
        """Draw the healpix map.

        Parameters
        ----------
        metric_value : `numpy.ma.MaskedArray`
            The metric values from the bundle.
        slicer : `rubin_sim.maf.slicers.TwoDSlicer`
            The slicer
        """
        kwargs = {}
        kwargs.update(self.default_hpixmap_kwargs)

        # Check if we have a valid HEALpix slicer
        if "Heal" not in slicer.slicer_name:
            hpix_map = self._transform_to_healpix(metric_values_in, slicer)
        else:
            hpix_map = metric_values_in.copy()

        hpix_map = apply_zp_norm(hpix_map, self.plot_dict)

        if "draw_hpxmap_kwargs" in self.plot_dict:
            kwargs.update(self.plot_dict["draw_hpxmap_kwargs"])

        try:
            kwargs["cmap"] = set_color_map(self.plot_dict)
        except AttributeError:
            # Probably here because an invalid cmap was set
            pass

        if self.plot_dict["log_scale"]:
            if "norm" in kwargs and kwargs["norm"] != "log":
                raise ValueError("Contradictory values for color scale normalization set.")
            kwargs["norm"] = "log"

        clims = set_color_lims(hpix_map, self.plot_dict)
        if "norm" not in kwargs or self._compatible_color_limits(kwargs["norm"], clims):
            kwargs["vmin"] = clims[0]
            kwargs["vmax"] = clims[1]
        else:
            warnings.warn(
                "Using log normalization, but color limits pass through 0. "
                "Adjusting so plotting doesn't fail"
            )

        # Filling in masked values must take place after automatic color limits
        # might be calculated!

        # This masking replicates the behavior of HealpixSkyMap, but is it
        # what we really want to do? Why not use the default handling of
        # masked values applied by skyproj?
        if "mask_below" in self.plot_dict and self.plot_dict["mask_below"] is not None:
            to_mask = np.where(hpix_map <= self.plot_dict["mask_below"])[0]
            hpix_map.mask[to_mask] = True
            hpix_map = hpix_map.filled(hp.UNSEEN)
        else:
            hpix_map = hpix_map.filled(slicer.badval)

        self.skyproj.draw_hpxmap(hpix_map, **kwargs)


class VisitPerimeterPlotter(SkyprojPlotter):
    default_visits_kwargs = {
        "edgecolor": "black",
        "linewidth": 0.2,
    }

    def __init__(self):
        super().__init__()
        self.plot_type = "SkyprojVisitPerimeter"
        self.object_plotter = True

    def draw(self, metric_values, slicer):
        """Draw the perimeters of visits.

        Parameters
        ----------
        metric_value : `numpy.ma.MaskedArray`
            The metric values from the bundle.
        slicer : `rubin_sim.maf.slicers.TwoDSlicer`
            The slicer
        """

        # If there are no visits, just return
        if metric_values is None or len(metric_values) == 0:
            return

        kwargs = {}
        kwargs.update(self.default_visits_kwargs)

        visits = metric_values[0]

        if "draw_polygon_kwargs" in self.plot_dict:
            kwargs.update(self.plot_dict["draw_polygon_kwargs"])

        camera_perimeter_func = self.plot_dict["camera_perimeter_func"]

        ras, decls = camera_perimeter_func(visits["fieldRA"], visits["fieldDec"], visits["rotSkyPos"])
        for visit_ras, visit_decls in zip(ras, decls):
            self.skyproj.draw_polygon(visit_ras, visit_decls, **kwargs)



from .altaz_batch import *
from .col_map_dict import *
from .common import *
from .ddf_batch import *
from .filterchange_batch import *
from .glance_batch import *
from .hourglass_batch import *
from .info_batch import *
from .metadata_batch import *
from .moving_objects_batch import *
from .openshutter_batch import *
from .radar_limited import *
from .science_radar_batch import *
from .skycoverage import *
from .slew_batch import *
from .srd_batch import *
from .time_batch import *
from .visitdepth_batch import *



# Example for IntervalsBetweenObsMetric
# Somayeh Khakpash - Lehigh University
# Last edited : 10/21/2020
# Calculates statistics (mean or median or standard deviation) of intervals
# between observations during simultaneous windows/Inter-seasonal gap of
# another survey.
# SurveyIntervals is the list of the survey observing window/Inter-seasonal
# gap intervals. It should be in the format:
# SurveyIntervals = [ [YYYY-MM-DD, YYYY-MM-DD] , [YYYY-MM-DD, YYYY-MM-DD] ,
# ... , [YYYY-MM-DD, YYYY-MM-DD] ]
# We are interested in calculating this metric in each of the LSST passbands.
# The difference between this metric and the VisitGapMetric metric is that
# VisitGapMetric calculates reduceFunc of gaps between observations of a
# data_slice throughout the whole
# baseline, but IntervalsBetweenObsMetric calculates the gaps between
# observations during another survey observing window.
# This metric combined with surveys footprint
# overlap can determine how many often another survey footprint is
# observed by LSST during specific time intervals.
__all__ = ("IntervalsBetweenObsMetric",)

import numpy as np
from astropy.time import Time

from rubin_sim.maf.metrics import BaseMetric


class IntervalsBetweenObsMetric(BaseMetric):
    def __init__(
        self,
        survey_intervals,
        stat,
        metric_name="IntervalsBetweenObsMetric",
        time_col="observationStartMJD",
        **kwargs,
    ):
        self.time_col = time_col
        self.metric_name = metric_name
        self.survey_intervals = survey_intervals
        self.stat = stat
        super(IntervalsBetweenObsMetric, self).__init__(col=time_col, metric_name=metric_name, **kwargs)

    def run(self, data_slice, slice_point=None):
        data_slice.sort(order=self.time_col)
        obs_diff = []

        for interval in self.survey_intervals:
            start_interval = Time(interval[0] + " 00:00:00")
            end_interval = Time(interval[1] + " 00:00:00")
            index = data_slice[self.time_col][
                np.where(
                    (data_slice[self.time_col] > start_interval.mjd)
                    & (data_slice[self.time_col] < end_interval.mjd)
                )[0]
            ]
            obs_diff_per_interval = np.diff(index)
            obs_diff = obs_diff + obs_diff_per_interval.tolist()

        if self.stat == "mean":
            result = np.mean(obs_diff)

        elif self.stat == "median":
            result = np.median(obs_diff)

        elif self.stat == "std":
            result = np.std(obs_diff)

        return result



# imports
import unittest

import numpy as np
from matplotlib.figure import Figure
from rubin_scheduler.scheduler.model_observatory import ModelObservatory

import rubin_sim.maf as maf


class TestPlotters(unittest.TestCase):
    def setUp(self):
        # Set a seed to make the tests reproducible
        self.rng = np.random.default_rng(seed=1234)

    def test_healpix_plotters(self):
        # Set up a metric bundle to send to plotters
        bundle1 = maf.create_empty_metric_bundle()
        nside = 64
        bundle1.slicer = maf.HealpixSlicer(nside=nside)
        bundle1._setup_metric_values()
        bundle1.metric_values += self.rng.uniform(size=len(bundle1.slicer))
        # First test healpix sky map - just that it runs.
        bundle1.set_plot_funcs([maf.HealpixSkyMap()])
        figs = bundle1.plot()
        self.assertTrue(isinstance(figs["SkyMap"], Figure))
        # Test healpix histogram - just that it runs
        bundle1.set_plot_funcs([maf.HealpixHistogram()])
        figs = bundle1.plot()
        self.assertTrue(isinstance(figs["Histogram"], Figure))
        # Test power spectrum
        bundle1.set_plot_funcs([maf.HealpixPowerSpectrum()])
        figs = bundle1.plot()
        self.assertTrue(isinstance(figs["PowerSpectrum"], Figure))

    def test_base_skymap(self):
        bundle1 = maf.create_empty_metric_bundle()
        npoints = 1000
        ra = self.rng.uniform(low=0, high=360, size=npoints)
        dec = self.rng.uniform(low=-90, high=90, size=npoints)
        bundle1.slicer = maf.UserPointsSlicer(ra, dec)
        bundle1._setup_metric_values()
        bundle1.metric_values += self.rng.uniform(size=len(bundle1.slicer))
        # Test skymap
        bundle1.set_plot_funcs([maf.BaseSkyMap()])
        figs = bundle1.plot()
        self.assertTrue(isinstance(figs["SkyMap"], Figure))
        # Test healpix histogram - just that it runs
        bundle1.set_plot_funcs([maf.BaseHistogram()])
        figs = bundle1.plot()
        self.assertTrue(isinstance(figs["Histogram"], Figure))

    def test_oned_plotter(self):
        bundle1 = maf.create_empty_metric_bundle()
        npoints = 100
        bins = np.arange(0, npoints, 1)
        bundle1.slicer = maf.OneDSlicer(slice_col_name="test", bins=bins)
        bundle1.slicer.slice_points = {"bins": bins}
        bundle1.slicer.nslice = len(bins) - 1
        bundle1.slicer.shape = len(bins) - 1
        bundle1._setup_metric_values()
        bundle1.metric_values += self.rng.uniform(size=len(bundle1.slicer))
        # Test plotter
        bundle1.set_plot_funcs([maf.OneDBinnedData()])
        figs = bundle1.plot()
        self.assertTrue(isinstance(figs["BinnedData"], Figure))

    def test_hpxmap_plotter(self):
        bundle = maf.create_empty_metric_bundle()
        nside = 64
        bundle.slicer = maf.HealpixSlicer(nside=nside)
        bundle._setup_metric_values()
        bundle.metric_values += self.rng.uniform(size=len(bundle.slicer))

        plotter = maf.HpxmapPlotter()
        bundle.set_plot_funcs([plotter])
        _ = bundle.plot()

    def test_visit_perimeter_plotter(self):
        model_observatory = ModelObservatory(init_load_length=1)

        num_points = 5
        field_ra = np.arange(30, 30 + num_points, dtype=float)
        field_dec = np.arange(-60, -60 + num_points, dtype=float)
        rot_sky_pos = np.arange(num_points, dtype=float) % 360

        unmasked_data = np.empty(dtype=object, shape=(1,))
        unmasked_data[0] = np.core.records.fromarrays(
            (field_ra, field_dec, rot_sky_pos),
            dtype=np.dtype(
                [
                    ("fieldRA", field_ra.dtype),
                    ("fieldDec", field_dec.dtype),
                    ("rotSkyPos", rot_sky_pos.dtype),
                ]
            ),
        )
        masked_data = np.ma.MaskedArray(data=unmasked_data, mask=False, fill_value=-666, dtype=object)

        bundle = maf.create_empty_metric_bundle()
        bundle.slicer = maf.UniSlicer()
        bundle._setup_metric_values()
        bundle.metric_values = masked_data

        def compute_camera_perimeter(ra, decl, rotation):
            # just a quadrangle for this unit test.
            # the math isn't quite right for an actual square,
            # but good enough for a unit test.
            size = 1.0
            ras = [
                ra - 0.5 * size * np.cos(np.radians(decl)) * np.sin(np.radians(rotation)),
                ra - 0.5 * size * np.cos(np.radians(decl)) * np.cos(np.radians(rotation)),
                ra + 0.5 * size * np.cos(np.radians(decl)) * np.sin(np.radians(rotation)),
                ra + 0.5 * size * np.cos(np.radians(decl)) * np.cos(np.radians(rotation)),
            ]
            decls = [
                decl + 0.5 * size * np.cos(np.radians(rotation)),
                decl - 0.5 * size * np.sin(np.radians(rotation)),
                decl - 0.5 * size * np.cos(np.radians(rotation)),
                decl + 0.5 * size * np.sin(np.radians(rotation)),
            ]
            return ras, decls

        plot_dict = {
            "camera_perimeter_func": compute_camera_perimeter,
            "model_observatory": model_observatory,
            "decorations": ["ecliptic", "galactic_plane", "sun", "moon", "horizon"],
        }
        plotter = maf.VisitPerimeterPlotter()
        bundle.set_plot_funcs([plotter])
        bundle.set_plot_dict(plot_dict)
        _ = bundle.plot()


if __name__ == "__main__":
    unittest.main()



# imports
import os
import sys
import unittest
from os import path
from tempfile import TemporaryDirectory
from unittest.mock import patch

import pandas as pd

from rubin_sim import maf
from rubin_sim.data import get_data_dir

# constants

URLROOT = "https://raw.githubusercontent.com/lsst-pst/survey_strategy/main/fbs_2.0/"
FAMILY_SOURCE = URLROOT + "runs_v2.1.json"
METRIC_SUBSET_SOURCE = os.path.join(get_data_dir(), "maf", "metric_subsets.json")
SUMMARY_SOURCE = os.path.join(get_data_dir(), "maf", "baseline_summary.h5")

# exception classes

# interface functions

# classes


class TestArchive(unittest.TestCase):
    num_runs = 11
    num_metrics = 7

    def test_get_family_runs(self):
        runs = maf.get_family_runs(FAMILY_SOURCE)

        self.assertIsInstance(runs, pd.DataFrame)

        columns = ("run", "brief", "filepath", "url")
        for column in columns:
            self.assertIn(column, runs.columns)

        self.assertEqual(runs.index.name, "family")

    def test_get_runs(self):
        runs = maf.get_runs(FAMILY_SOURCE)
        self.assertEqual(runs.index.name, "run")

        columns = ("family", "brief", "filepath", "url")
        for column in columns:
            self.assertIn(column, runs.columns)

        self.assertEqual(runs.index.name, "run")

    @unittest.skip("skipping long running test.")
    def test_download_runs(self):
        run = maf.get_runs(FAMILY_SOURCE).index.values[0]

        temp_dir_itself = TemporaryDirectory()
        temp_dir = temp_dir_itself.name

        dest_fnames = maf.download_runs(run, dest_dir=temp_dir)
        self.assertTrue(path.exists(dest_fnames[0]))

        temp_dir_itself.cleanup()

    def test_get_metric_subsets(self):
        metric_subsets = maf.get_metric_subsets(METRIC_SUBSET_SOURCE)
        self.assertIsInstance(metric_subsets, pd.DataFrame)
        self.assertIn("metric subset", metric_subsets.index.names)
        self.assertIn("metric", metric_subsets.index.names)

        columns = ("style", "invert", "mag")
        for column in columns:
            self.assertIn(column, metric_subsets.columns)

    def test_get_metric_summaries(self):
        summary = maf.get_metric_summaries(summary_source=SUMMARY_SOURCE)
        self.assertIsInstance(summary, pd.DataFrame)
        self.assertEqual(summary.columns.name, "metric")
        self.assertEqual(summary.index.name, "run")

        my_summary = maf.get_metric_summaries(
            runs=["baseline_v3.0_10yrs", "baseline_v3.4_10yrs"],
            metrics=[
                "Rms Max normairmass All sky all bands HealpixSlicer",
                "Median Max normairmass All sky all bands HealpixSlicer",
                "Max Max normairmass All sky all bands HealpixSlicer",
            ],
            summary_source=summary,
        )
        self.assertEqual(len(my_summary), 2)
        self.assertEqual(len(my_summary.columns), 3)

        srd_sim = maf.get_metric_summaries(
            metric_subsets="SRD",
            summary_source=summary,
            metric_subset_source=METRIC_SUBSET_SOURCE,
        )
        self.assertGreater(len(srd_sim.columns), 3)
        self.assertLess(len(srd_sim.columns), len(summary.columns))

        srdnvis_sim = maf.get_metric_summaries(
            metric_subsets=["SRD", "N Visits"],
            summary_source=summary,
            metric_subset_source=METRIC_SUBSET_SOURCE,
        )
        self.assertGreater(len(srdnvis_sim.columns), len(srd_sim.columns))
        self.assertLess(len(srdnvis_sim.columns), len(summary.columns))

    def test_get_family_descriptions(self):
        families = maf.get_family_descriptions(FAMILY_SOURCE)
        self.assertIsInstance(families, pd.DataFrame)
        self.assertEqual(families.index.name, "family")

    def test_describe_families(self):
        if "IPython" in sys.modules:
            with (
                patch("IPython.display.display_markdown") as _,
                patch("IPython.display.HTML") as _,
                patch("IPython.display") as _,
            ):
                self.perform_describe_families_test()
        else:
            with patch("builtins.print") as _:
                self.perform_describe_families_test()

    @unittest.skip("Skipping; family descriptions out of sync with summary")
    def perform_describe_families_test(self):
        families = maf.get_family_descriptions(FAMILY_SOURCE)
        disp_families = families[:2]

        fig, ax = maf.describe_families(families)
        self.assertIsNone(fig)
        self.assertIsNone(ax)

        all_metric_subsets = maf.get_metric_subsets(METRIC_SUBSET_SOURCE)
        summary = maf.get_metric_summaries(summary_source=SUMMARY_SOURCE)

        table_metric_subset = all_metric_subsets.loc["SRD"]
        fig, ax = maf.describe_families(
            disp_families, summary=summary, table_metric_subset=table_metric_subset
        )
        self.assertIsNone(fig)
        self.assertIsNone(ax)

        plot_metric_subset = all_metric_subsets.loc["N Visits"]
        fig, ax = maf.describe_families(disp_families, summary=summary, plot_metric_subset=plot_metric_subset)

    def test_create_metric_subset(self):
        metrics = ["Urania", "Thalia", "Calliope", "Terpsichore"]
        metric_subset_name = "Muses"
        metric_subset = maf.create_metric_subset(metric_subset_name, metrics)
        self.assertSequenceEqual(metrics, metric_subset.metric.tolist())
        self.assertSequenceEqual(
            metric_subset.columns.tolist(),
            ["metric", "short_name", "style", "invert", "mag"],
        )
        self.assertSequenceEqual(metric_subset.index.names, ["metric subset", "metric"])


run_tests_now = __name__ == "__main__"
if run_tests_now:
    unittest.main()



from .bandpass import *
from .photometric_parameters import *
from .physical_parameters import *
from .predicted_zeropoints import *
from .sed import *
from .signaltonoise import *
from .spectral_resampling import *



import matplotlib.pyplot as plt
import numpy as np
import numpy.lib.recfunctions as rf
from scipy import interpolate


class Lims:
    """class to handle light curve of SN

    Parameters
    -------------
    Li_files : `str`
       light curve reference file
    mag_to_flux_files : `str`
       files of magnitude to flux
    band : `str`
        band considered
    SNR : `float`
        Signal-To-Noise Ratio cut
    mag_range : `float`, `float`, optional
        mag range considered
        Default : (23., 27.5)
    dt_range : `float`, `float`, optional
        difference time range considered (cadence)
        Default : (0.5, 25.)
    """

    def __init__(
        self,
        li_files,
        mag_to_flux_files,
        band,
        SNR,
        mag_range=(23.0, 27.5),
        dt_range=(0.5, 25.0),
    ):
        self.band = band
        self.SNR = SNR
        self.lims = []
        self.mag_to_flux = []
        self.mag_range = mag_range
        self.dt_range = dt_range

        for val in li_files:
            self.lims.append(self.get_lims(self.band, np.load(val), SNR))
        for val in mag_to_flux_files:
            self.mag_to_flux.append(np.load(val))
        self.interp()

    def get_lims(self, band, tab, SNR):
        """
        Estimations of the limits

        Parameters
        -------------
        band : `str`
          band to consider
        tab : `np.ndarray`, (N,)
          table of data
        SNR : `float`
           Signal-to-Noise Ratio cut

        Returns
        ---------
        lims : `dict`
            dict of limits with redshift and band as keys.

        """

        lims = {}

        for z in np.unique(tab["z"]):
            idx = (tab["z"] == z) & (tab["band"] == "LSST::" + band)
            idx &= tab["flux_e"] > 0.0
            sel = tab[idx]

            if len(sel) > 0:
                li2 = np.sqrt(np.sum(sel["flux_e"] ** 2))
                lim = 5.0 * li2 / SNR
                if z not in lims.keys():
                    lims[z] = {}
                lims[z][band] = lim

        return lims

    def mesh(self, mag_to_flux):
        """Mesh grid to estimate five-sigma depth values (m5) from mags input.

        Parameters
        -----------
        mag_to_flux : `np.ndarray`, (N,2)
            magnitude to flux values

        Returns
        -------
        m5, Dt, metric : `np.ndarray`, np.ndarray`, `np.ndarray`
            m5 values, time difference dt (cadence),
            metric=sqrt(dt)*f5 where f5 is the 5-sigma flux

        """
        dt = np.linspace(self.dt_range[0], self.dt_range[1], 100)
        m5 = np.linspace(self.mag_range[0], self.mag_range[1], 50)
        ida = mag_to_flux["band"] == self.band
        fa = interpolate.interp1d(mag_to_flux[ida]["m5"], mag_to_flux[ida]["flux_e"])
        f5 = fa(m5)
        f5, DT = np.meshgrid(f5, dt)
        m5, DT = np.meshgrid(m5, dt)
        metric = np.sqrt(DT) * f5

        return m5, DT, metric

    def interp(self):
        """Estimate a grid of interpolated values in the plane
        (m5, cadence, metric).
        """

        m5_all = []
        dt_all = []
        metric_all = []

        for val in self.mag_to_flux:
            m5, DT, metric = self.mesh(val)
            m5_all.append(m5)
            dt_all.append(DT)
            metric_all.append(metric)

        sorted_keys = []
        for i in range(len(self.lims)):
            sorted_keys.append(np.sort([k for k in self.lims[i].keys()])[::-1])
        figa, axa = plt.subplots()

        for kk, lim in enumerate(self.lims):
            ll = [lim[zz][self.band] for zz in sorted_keys[kk]]
            cs = axa.contour(m5_all[kk], dt_all[kk], metric_all[kk], ll)

            points_values = None
            for io, col in enumerate(cs.collections):
                # Update in matplotlib changed get_segments to get_paths
                if hasattr(col, "get_segments"):
                    segments = col.get_segments()
                else:
                    segments = col.get_paths()
                if segments:
                    segments = segments[0]
                    if hasattr(segments, "vertices"):
                        segments = segments.vertices
                    myarray = segments
                    res = np.array(myarray[:, 0], dtype=[("m5", "f8")])
                    res = rf.append_fields(res, "cadence", myarray[:, 1])
                    res = rf.append_fields(res, "z", [sorted_keys[kk][io]] * len(res))
                    if points_values is None:
                        points_values = res
                    else:
                        points_values = np.concatenate((points_values, res))
            self.points_ref = points_values

        plt.close(figa)  # do not display

    def interp_griddata(self, data):
        """Estimate metric interpolation for data (m5,cadence)

        Parameters
        ----------
        data : `np.ndarray`
            data where interpolation has to be done (m5,cadence)

        Returns
        --------
        res : `np.ndarray`
            griddata interpolation (m5,cadence,metric)
        """

        ref_points = self.points_ref
        res = interpolate.griddata(
            (ref_points["m5"], ref_points["cadence"]),
            ref_points["z"],
            (data["m5_mean"], data["cadence_mean"]),
            method="cubic",
        )
        return res


class GenerateFakeObservations:
    """Class to generate Fake observations

    Parameters
    -----------
    config: yaml-like
       configuration file (parameter choice: filter, cadence, m5,Nseasons, ..)
    list : `str`, optional
        Name of the columns used.
        Default : 'observationStartMJD', 'fieldRA', 'fieldDec','filter',
        'fiveSigmaDepth', 'visitExposureTime','numExposures',
        'visitTime','season'

    Returns
    ---------
    observations : `np.ndarray`, (N, M)
        recordarray of observations with the fields:
        MJD, Ra, Dec, band,m5,Nexp, ExpTime, Season
    """

    def __init__(
        self,
        config,
        mjd_col="observationStartMJD",
        ra_col="fieldRA",
        dec_col="fieldDec",
        filter_col="filter",
        m5_col="fiveSigmaDepth",
        exptime_col="visitExposureTime",
        nexp_col="numExposures",
        season_col="season",
    ):
        self.mjd_col = mjd_col
        self.m5_col = m5_col
        self.filter_col = filter_col
        self.ra_col = ra_col
        self.dec_col = dec_col
        self.exptime_col = exptime_col
        self.season_col = season_col
        self.nexp_col = nexp_col

        # now make fake obs
        self.make_fake(config)

    def make_fake(self, config):
        """Generate Fake observations

        Parameters
        -----------
        config: yaml-like
            configuration file (parameter choice: filter,
            cadence, m5,Nseasons, ...)
        """
        bands = config["bands"]
        cadence = dict(zip(bands, config["Cadence"]))
        shift_days = dict(zip(bands, [config["shift_days"] * io for io in range(len(bands))]))
        m5 = dict(zip(bands, config["m5"]))
        nvisits = dict(zip(bands, config["nvisits"]))
        exposure__time = dict(zip(bands, config["exposure__time"]))

        ra = config["ra"]
        dec = config["dec"]
        rtot = []
        # for season in range(1, config['nseasons']+1):
        for il, season in enumerate(config["seasons"]):
            # mjd_min = config['MJD_min'] + float(season-1)*inter_season_gap
            mjd_min = config["MJD_min"][il]
            mjd_max = mjd_min + config["season_length"]

            for i, band in enumerate(bands):
                mjd = np.arange(mjd_min, mjd_max + cadence[band], cadence[band])
                mjd += shift_days[band]
                m5_coadded = self.m5_coadd(m5[band], nvisits[band], exposure__time[band])
                myarr = np.array(mjd, dtype=[(self.mjd_col, "f8")])
                myarr = rf.append_fields(
                    myarr,
                    [self.ra_col, self.dec_col, self.filter_col],
                    [[ra] * len(myarr), [dec] * len(myarr), [band] * len(myarr)],
                )
                myarr = rf.append_fields(
                    myarr,
                    [self.m5_col, self.nexp_col, self.exptime_col, self.season_col],
                    [
                        [m5_coadded] * len(myarr),
                        [nvisits[band]] * len(myarr),
                        [nvisits[band] * exposure__time[band]] * len(myarr),
                        [season] * len(myarr),
                    ],
                )
                rtot.append(myarr)

        res = np.copy(np.concatenate(rtot))
        res.sort(order=self.mjd_col)

        self.observations = res

    def m5_coadd(self, m5, nvisits, tvisit):
        """Coadded m5 estimation

        Parameters
        ----------
        m5 : `list` [`float`]
           list of five-sigma depth values
        nvisits : `list` [`float`]
            list of the number of visits
        tvisit : `list` [`float`]
           list of the visit times

        Returns
        ---------
        m5_coadd : `list` [`float`]
            list of m5 coadded values
        """
        m5_coadd = m5 + 1.25 * np.log10(float(nvisits) * tvisit / 30.0)
        return m5_coadd


class ReferenceData:
    """class to handle light curve of SN

    Parameters
    ------------
    Li_files : `str`
        light curve reference file
    mag_to_flux_files : `str`
        files of magnitude to flux
    band : `str`
        band considered
    z : `float`
        redshift considered
    """

    def __init__(self, li_files, mag_to_flux_files, band, z):
        self.band = band
        self.z = z
        self.fluxes = []
        self.mag_to_flux = []

        for val in li_files:
            self.fluxes.append(self.interp_fluxes(self.band, np.load(val), self.z))
        for val in mag_to_flux_files:
            self.mag_to_flux.append(self.interp_mag(self.band, np.load(val)))

    def interp_fluxes(self, band, tab, z):
        """Flux interpolator

        Parameters
        ---------------
        band : `str`
            band considered
        tab : `np.ndarray`
            reference data with (at least) fields z,band,time,DayMax
        z : `float`
            redshift considered

        Returns
        --------
        fluxes : `list` [`float`]
            list (float) of interpolated fluxes (in e/sec)
        """
        idx = (np.abs(tab["z"] - z) < 1.0e-5) & (tab["band"] == "LSST::" + band)
        sel = tab[idx]
        selc = np.copy(sel)
        difftime = sel["time"] - sel["DayMax"]
        selc = rf.append_fields(selc, "deltaT", difftime)
        return interpolate.interp1d(selc["deltaT"], selc["flux_e"], bounds_error=False, fill_value=0.0)

    def interp_mag(self, band, tab):
        """magnitude (m5) to flux (e/sec) interpolator

        Parameters
        ---------------
        band : `str`
            band considered
        tab : `np.ndarray`
            reference data with (at least) fields band,m5,flux_e,
        z : `float`
            redshift considered

        Returns
        --------
        mags : `list` [`float`]
            list (float) of interpolated magnitudes (in e/sec)
        """
        idx = tab["band"] == band
        sel = tab[idx]
        return interpolate.interp1d(sel["m5"], sel["flux_e"], bounds_error=False, fill_value=0.0)



import os
import unittest

import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal
from rubin_scheduler.data import get_data_dir

from rubin_sim.moving_objects import Orbits


class TestOrbits(unittest.TestCase):
    def setUp(self):
        self.testdir = os.path.join(get_data_dir(), "tests", "orbits_testdata")

    def test_equal_not_equal(self):
        orbits = Orbits()
        orbits.read_orbits(os.path.join(self.testdir, "test_orbitsQ.des"))
        self.assertEqual(len(orbits), 4)
        orbits2 = Orbits()
        orbits2.read_orbits(os.path.join(self.testdir, "test_orbitsQ.des"))
        self.assertEqual(orbits, orbits2)
        orbits3 = Orbits()
        orbits3.read_orbits(os.path.join(self.testdir, "test_orbitsA.des"))
        self.assertNotEqual(orbits, orbits3)

    def test_iteration_and_indexing(self):
        orbits = Orbits()
        orbits.read_orbits(os.path.join(self.testdir, "test_orbitsNEO.s3m"))
        orbits_single = orbits[0]
        assert_frame_equal(orbits_single.orbits, orbits.orbits.query("index==0"))
        orbits_single = orbits[3]
        assert_frame_equal(orbits_single.orbits, orbits.orbits.query("index==3"))
        # Test iteration through all orbits.
        for orb, (i, orbi) in zip(orbits, orbits.orbits.iterrows()):
            self.assertEqual(orb.orbits.obj_id.values[0], orbi.obj_id)
            self.assertTrue(isinstance(orb, Orbits))
            self.assertEqual(orb.orbits.index, i)
        # Test iteration through a subset of orbits.
        orbits_sub = Orbits()
        orbits_sub.set_orbits(orbits.orbits.query("index > 4"))
        for orb, (i, orbi) in zip(orbits_sub, orbits_sub.orbits.iterrows()):
            self.assertEqual(orb.orbits.obj_id.values[0], orbi.obj_id)
            self.assertTrue(isinstance(orb, Orbits))
            self.assertEqual(orb.orbits.index, i)

    def test_slicing(self):
        """
        Test that we can slice a collection of orbits
        """
        orbits = Orbits()
        orbits.read_orbits(os.path.join(self.testdir, "test_orbitsNEO.s3m"))
        orbit_slice = orbits[2:6]
        self.assertEqual(orbit_slice[0], orbits[2])
        self.assertEqual(orbit_slice[1], orbits[3])
        self.assertEqual(orbit_slice[2], orbits[4])
        self.assertEqual(orbit_slice[3], orbits[5])
        self.assertEqual(len(orbit_slice), 4)

        orbit_slice = orbits[1:7:2]
        self.assertEqual(orbit_slice[0], orbits[1])
        self.assertEqual(orbit_slice[1], orbits[3])
        self.assertEqual(orbit_slice[2], orbits[5])
        self.assertEqual(len(orbit_slice), 3)

    def test_offset_dataframe(self):
        """
        Test that we can slice and iterate through an orbits
        dataframe that has already been sub-selected from another
        dataframe.
        """
        orbits0 = Orbits()
        orbits0.read_orbits(os.path.join(self.testdir, "test_orbitsNEO.s3m"))

        orbits_sub = Orbits()
        orbits_sub.set_orbits(orbits0.orbits.query("index>1"))

        self.assertEqual(len(orbits_sub), 6)

        orbit_slice = orbits_sub[2:6]
        self.assertEqual(orbit_slice[0], orbits_sub[2])
        self.assertEqual(orbit_slice[1], orbits_sub[3])
        self.assertEqual(orbit_slice[2], orbits_sub[4])
        self.assertEqual(orbit_slice[3], orbits_sub[5])
        self.assertEqual(len(orbit_slice), 4)

        orbit_slice = orbits_sub[1:5:2]
        self.assertEqual(orbit_slice[0], orbits_sub[1])
        self.assertEqual(orbit_slice[1], orbits_sub[3])
        self.assertEqual(len(orbit_slice), 2)

        for ii, oo in enumerate(orbits_sub):
            self.assertEqual(oo, orbits0[ii + 2])

    def test_read_orbits(self):
        orbits = Orbits()
        orbits.read_orbits(os.path.join(self.testdir, "test_orbitsQ.des"))
        self.assertEqual(len(orbits), 4)
        orbits.read_orbits(os.path.join(self.testdir, "test_orbitsA.des"))
        self.assertEqual(len(orbits), 4)
        orbits.read_orbits(os.path.join(self.testdir, "test_orbitsCAR.des"))
        self.assertEqual(len(orbits), 1)
        with self.assertRaises(ValueError):
            orbits.read_orbits(os.path.join(self.testdir, "test_orbitsBadMix.des"))
        with self.assertRaises(ValueError):
            orbits.read_orbits(os.path.join(self.testdir, "test_orbitsBad.des"))

    def test_set_orbits(self):
        orbits = Orbits()
        orbits.read_orbits(os.path.join(self.testdir, "test_orbitsQ.des"))
        # Test that we can set the orbits using a dataframe.
        suborbits = orbits.orbits.head(1)
        new_orbits = Orbits()
        new_orbits.set_orbits(suborbits)
        self.assertEqual(len(new_orbits), 1)
        self.assertEqual(new_orbits.orb_format, "COM")
        assert_frame_equal(new_orbits.orbits, suborbits)
        # Test that we can set the orbits using a Series.
        for i, sso in suborbits.iterrows():
            new_orbits = Orbits()
            new_orbits.set_orbits(sso)
            self.assertEqual(len(new_orbits), 1)
            self.assertEqual(new_orbits.orb_format, "COM")
            assert_frame_equal(new_orbits.orbits, suborbits)
        # Test that we can set the orbits using a numpy array of many objects.
        numpyorbits = orbits.orbits.to_records(index=False)
        new_orbits = Orbits()
        new_orbits.set_orbits(numpyorbits)
        self.assertEqual(len(new_orbits), len(orbits))
        self.assertEqual(new_orbits.orb_format, "COM")
        assert_frame_equal(new_orbits.orbits, orbits.orbits)
        # And test that this works for a single row of the numpy array.
        onenumpyorbits = numpyorbits[0]
        new_orbits = Orbits()
        new_orbits.set_orbits(onenumpyorbits)
        self.assertEqual(len(new_orbits), 1)
        self.assertEqual(new_orbits.orb_format, "COM")
        assert_frame_equal(new_orbits.orbits, suborbits)
        # And test that it fails appropriately when columns are not correct.
        neworbits = pd.DataFrame(orbits.orbits)
        newcols = neworbits.columns.values.tolist()
        newcols[0] = "ssmId"
        newcols[3] = "ecc"
        neworbits.columns = newcols
        new_orbits = Orbits()
        with self.assertRaises(ValueError):
            new_orbits.set_orbits(neworbits)

    def test_set_seds(self):
        """
        Test that the self-assignment of SEDs works as expected.
        """
        orbits = Orbits()
        # Test with a range of a values.
        a = np.arange(0, 5, 0.05)
        orbs = pd.DataFrame(a, columns=["a"])
        seds = orbits.assign_sed(orbs)
        self.assertEqual(np.unique(seds[np.where(a < 2)]), "S.dat")
        self.assertEqual(np.unique(seds[np.where(a > 4)]), "C.dat")
        # Test when read a values.
        orbits.read_orbits(os.path.join(self.testdir, "test_orbitsA.des"))
        sedvals = orbits.assign_sed(orbits.orbits, random_seed=42)
        orbits2 = Orbits()
        orbits2.read_orbits(os.path.join(self.testdir, "test_orbitsQ.des"))
        sedvals2 = orbits2.assign_sed(orbits2.orbits, random_seed=42)
        np.testing.assert_array_equal(sedvals, sedvals2)


if __name__ == "__main__":
    unittest.main()



__all__ = ("PhysicalParameters",)


class PhysicalParameters:
    """
    Stores physical constants and other immutable parameters
    used by the sims_phot_utils code.
    """

    def __init__(self):
        self._lightspeed = 299792458.0  # speed of light, m/s
        self._planck = 6.626068e-27  # planck's constant, ergs*seconds
        self._nm2m = 1.00e-9  # nanometers to meters conversion m/nm
        self._ergsetc2jansky = 1.00e23  # erg/cm2/s/Hz to Jansky units (fnu)

    @property
    def lightspeed(self):
        """Speed of light in meters per second."""
        return self._lightspeed

    @lightspeed.setter
    def lightspeed(self, value):
        raise RuntimeError("Cannot change the value of lightspeed " + "(Einstein does not approve)")

    @property
    def nm2m(self):
        """Conversion factor to go from nm to m."""
        return self._nm2m

    @nm2m.setter
    def nm2m(self, value):
        raise RuntimeError("Cannot change the value of nm2m")

    @property
    def ergsetc2jansky(self):
        """Conversion factor to go from ergs/sec/cm^2 to Janskys."""
        return self._ergsetc2jansky

    @ergsetc2jansky.setter
    def ergsetc2jansky(self, value):
        raise RuntimeError("Cannot change the value of ergsetc2Jansky")

    @property
    def planck(self):
        """Planck's constant in ergs*seconds."""
        return self._planck

    @planck.setter
    def planck(self, value):
        raise RuntimeError("Cannot change the value of planck")



__all__ = (
    "get_sim_data",
    "scale_benchmarks",
    "calc_coadded_depth",
)

import os
import sqlite3

import numpy as np
import pandas as pd
from sqlalchemy import create_engine, inspect
from sqlalchemy.engine import make_url


def get_sim_data(
    db_con,
    sqlconstraint,
    dbcols,
    stackers=None,
    table_name=None,
    full_sql_query=None,
):
    """Query an opsim database for the needed data columns
    and run any required stackers.

    Parameters
    ----------
    db_con : `str` or SQLAlchemy connectable, or sqlite3 connection
        Filename to a sqlite3 file, or a connection object that
        can be used by pandas.read_sql
    sqlconstraint : `str` or None
        SQL constraint to apply to query for observations.
        Ignored if full_sql_query is set.
    dbcols : `list` [`str`]
        Columns required from the database. Ignored if full_sql_query is set.
    stackers : `list` [`rubin_sim.maf.stackers`], optional
        Stackers to be used to generate additional columns. Default None.
    table_name : `str` (None)
        Name of the table to query.
        Default None will try "observations".
        Ignored if full_sql_query is set.
    full_sql_query : `str`
        The full SQL query to use. Overrides sqlconstraint, dbcols, tablename.

    Returns
    -------
    sim_data : `np.ndarray`
        A numpy structured array with columns resulting from dbcols + stackers,
        for observations matching the SQLconstraint.
    """
    if sqlconstraint is None:
        sqlconstraint = ""

    # Check that file exists
    if isinstance(db_con, str):
        if os.path.isfile(db_con) is False:
            raise FileNotFoundError("No file %s" % db_con)

    # Check if table is "observations" or "SummaryAllProps"
    if (table_name is None) & (full_sql_query is None) & (isinstance(db_con, str)):
        url = make_url("sqlite:///" + db_con)
        eng = create_engine(url)
        inspector = inspect(eng)
        tables = [inspector.get_table_names(schema=schema) for schema in inspector.get_schema_names()]
        if "observations" in tables[0]:
            table_name = "observations"
        elif "SummaryAllProps" in tables[0]:
            table_name = "SummaryAllProps"
        elif "summary" in tables[0]:
            table_name = "summary"
        else:
            ValueError("Could not guess table_name, set with table_name or full_sql_query kwargs")
    elif (table_name is None) & (full_sql_query is None):
        # If someone passes in a connection object with an old table_name
        # things will fail
        # that's probably fine, keep people from getting fancy with old sims
        table_name = "observations"

    if isinstance(db_con, str):
        con = sqlite3.connect(db_con)
    else:
        con = db_con

    if full_sql_query is None:
        col_str = ""
        for colname in dbcols:
            col_str += colname + ", "
        col_str = col_str[0:-2] + " "

        query = "SELECT %s FROM %s" % (col_str, table_name)
        if len(sqlconstraint) > 0:
            query += " WHERE %s" % (sqlconstraint)
        query += ";"
        sim_data = pd.read_sql(query, con).to_records(index=False)

    else:
        query = full_sql_query
        sim_data = pd.read_sql(query, con).to_records(index=False)

    if len(sim_data) == 0:
        raise UserWarning("No data found matching sqlconstraint %s" % (sqlconstraint))
    # Now add the stacker columns.
    if stackers is not None:
        for s in stackers:
            sim_data = s.run(sim_data)
    return sim_data


def scale_benchmarks(run_length, benchmark="design"):
    """Set design and stretch values of the number of visits or
    area of the footprint or seeing/Fwhmeff/skybrightness and single visit
    depth (based on SRD values).
    Scales number of visits for the length of the run, relative to 10 years.

    Parameters
    ----------
    run_length : `float`
        The length (in years) of the run.
    benchmark : `str`
        design or stretch - which version of the SRD values to return.

    Returns
    -------
    benchmarks: `dict` of floats
       A dictionary containing the number of visits, area of footprint,
       seeing and FWHMeff values, skybrightness and single visit depth
       for either the design or stretch SRD values.
    """
    # Set baseline (default) numbers for the baseline survey length (10 years).
    baseline = 10.0

    design = {}
    stretch = {}

    design["nvisitsTotal"] = 825
    stretch["nvisitsTotal"] = 1000
    design["Area"] = 18000
    stretch["Area"] = 20000

    design["nvisits"] = {"u": 56, "g": 80, "r": 184, "i": 184, "z": 160, "y": 160}
    stretch["nvisits"] = {"u": 70, "g": 100, "r": 230, "i": 230, "z": 200, "y": 200}

    # mag/sq arcsec
    design["skybrightness"] = {
        "u": 21.8,
        "g": 22.0,
        "r": 21.3,
        "i": 20.0,
        "z": 19.1,
        "y": 17.5,
    }
    stretch["skybrightness"] = {
        "u": 21.8,
        "g": 22.0,
        "r": 21.3,
        "i": 20.0,
        "z": 19.1,
        "y": 17.5,
    }

    # arcsec - old seeing values
    design["seeing"] = {"u": 0.77, "g": 0.73, "r": 0.7, "i": 0.67, "z": 0.65, "y": 0.63}
    stretch["seeing"] = {
        "u": 0.77,
        "g": 0.73,
        "r": 0.7,
        "i": 0.67,
        "z": 0.65,
        "y": 0.63,
    }

    # arcsec - new FWHMeff values (scaled from old seeing)
    design["FWHMeff"] = {
        "u": 0.92,
        "g": 0.87,
        "r": 0.83,
        "i": 0.80,
        "z": 0.78,
        "y": 0.76,
    }
    stretch["FWHMeff"] = {
        "u": 0.92,
        "g": 0.87,
        "r": 0.83,
        "i": 0.80,
        "z": 0.78,
        "y": 0.76,
    }

    design["singleVisitDepth"] = {
        "u": 23.9,
        "g": 25.0,
        "r": 24.7,
        "i": 24.0,
        "z": 23.3,
        "y": 22.1,
    }
    stretch["singleVisitDepth"] = {
        "u": 24.0,
        "g": 25.1,
        "r": 24.8,
        "i": 24.1,
        "z": 23.4,
        "y": 22.2,
    }

    # Scale the number of visits.
    if run_length != baseline:
        scalefactor = float(run_length) / float(baseline)
        # Calculate scaled value for design and stretch values of nvisits,
        # per filter.
        for f in design["nvisits"]:
            design["nvisits"][f] = int(np.floor(design["nvisits"][f] * scalefactor))
            stretch["nvisits"][f] = int(np.floor(stretch["nvisits"][f] * scalefactor))

    if benchmark == "design":
        return design
    elif benchmark == "stretch":
        return stretch
    else:
        raise ValueError("Benchmark value %s not understood: use 'design' or 'stretch'" % (benchmark))


def calc_coadded_depth(nvisits, single_visit_depth):
    """Calculate the coadded depth expected for a given number of visits
    and single visit depth.

    Parameters
    ----------
    nvisits : `dict` of `int` or `float`
        Dictionary (per filter) of number of visits
    single_visit_depth : `dict` of `float`
        Dictionary (per filter) of the single visit depth

    Returns
    -------
    coadded_depth : `dict` of `float`
        Dictionary of coadded depths per filter.
    """
    coadded_depth = {}
    for f in nvisits:
        if f not in single_visit_depth:
            raise ValueError("Filter keys in nvisits and single_visit_depth must match")
        coadded_depth[f] = float(1.25 * np.log10(nvisits[f] * 10 ** (0.8 * single_visit_depth[f])))
        if not np.isfinite(coadded_depth[f]):
            coadded_depth[f] = single_visit_depth[f]
    return coadded_depth
